# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2001-2006  James Henstridge
#
#   darcs.py: some code to handle various darcs operations
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

__all__ = []
__metaclass__ = type

import os
import urlparse

try:
    import hashlib
except ImportError:
    import md5 as hashlib

from jhbuild.errors import FatalError, CommandError
from jhbuild.versioncontrol import Repository, Branch, register_repo_type
from jhbuild.commands.sanitycheck import inpath

class DarcsRepository(Repository):
    """A class representing a Darcs repository.

    Note that this is just the parent directory for a bunch of darcs
    branches, making it easy to switch to a mirror URI.
    """

    init_xml_attrs = ['href']

    def __init__(self, config, name, href):
        Repository.__init__(self, config, name)
        # allow user to adjust location of branch.
        self.href = config.repos.get(name, href)

    branch_xml_attrs = ['module', 'checkoutdir']

    def branch(self, name, module=None, checkoutdir=None):
        if name in self.config.branches:
            module = self.config.branches[name]
            if not module:
                raise FatalError(_('branch for %(name)s has wrong override, check your %(filename)s') % \
                                   {'name'     : name,
                                    'filename' : self.config.filename})
        else:
            if module is None:
                module = name
            module = urlparse.urljoin(self.href, module)
        return DarcsBranch(self, module, checkoutdir)


class DarcsBranch(Branch):
    """A class representing a Darcs branch."""

    def srcdir(self):
        if self.checkoutdir:
            return os.path.join(self.checkoutroot, self.checkoutdir)
        else:
            return os.path.join(self.checkoutroot,
                                os.path.basename(self.module))
    srcdir = property(srcdir)

    def branchname(self):
        return None
    branchname = property(branchname)

    def _checkout(self, buildscript):
        cmd = ['darcs', 'get', self.module]
        if self.checkoutdir:
            cmd.append(self.checkoutdir)

        if self.config.sticky_date:
            raise FatalError(_('date based checkout not yet supported\n'))

        buildscript.execute(cmd, 'darcs', cwd=self.checkoutroot)

    def _update(self, buildscript):
        if self.config.sticky_date:
            raise FatalError(_('date based checkout not yet supported\n'))
        buildscript.execute(['darcs', 'pull', '-a', '--no-set-default', self.module],
                            'darcs', cwd=self.srcdir)

    def _fix_permissions(self):
        # This is a hack to make the autogen.sh and/or configure
        # scripts executable.  This is needed because Darcs does not
        # version the executable bit.
        for filename in ['autogen.sh', 'configure']:
            path = os.path.join(self.srcdir, filename)
            try:
                stat = os.stat(path)
            except OSError as e:
                continue
            os.chmod(path, stat.st_mode | 0111)

    def checkout(self, buildscript):
        if not inpath('darcs', os.environ['PATH'].split(os.pathsep)):
            raise CommandError(_('%s not found') % 'darcs')
        Branch.checkout(self, buildscript)
        self._fix_permissions()

    def tree_id(self):
        # XXX: check with some darcs expert if there is not a command to get
        # this
        if not os.path.exists(self.srcdir):
            return None
        return hashlib.md5(file(os.path.join(self.srcdir, '_darcs', 'inventory')).read()).hexdigest()

register_repo_type('darcs', DarcsRepository)



# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2011  Red Hat, Inc.
#
#   trigger.py - Run scripts after packages are installed
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os
import sys 
import subprocess
import re

from jhbuild.utils import cmds

class Trigger(object):
    SUFFIX = '.trigger'
    def __init__(self, filepath):
        assert filepath.endswith(self.SUFFIX)
        self._rematches = []
        self._literal_matches = []
        self._executable = None
        self._file = filepath
        self.name = os.path.basename(filepath)[:-len(self.SUFFIX)]

        f = open(self._file)
        for line in f:
            key = '# IfExecutable: '
            if line.startswith(key):
                text = line[len(key):].strip()
                self._executable = text
                continue
            key = '# REMatch: '
            if line.startswith(key):
                text = line[len(key):].strip()
                r = re.compile(text)
                self._rematches.append(r)
                continue
            key = '# LiteralMatch: '
            if line.startswith(key):
                text = line[len(key):].strip()
                self._literal_matches.append(text)
                continue
        f.close()
        if len(self._rematches) == 0 and len(self._literal_matches) == 0:
            raise ValueError(_("No keys specified in trigger script %r") % (filepath, ))
        
    def matches(self, files_list):
       """@files_list should be a list of absolute file paths.  Return True if this trigger script
should be run."""
       if self._executable is not None:
           if not cmds.has_command(self._executable):
               return False
       for path in files_list:
           for r in self._rematches:
               match = r.search(path)
               if match:
                   return True
           for literal in self._literal_matches:
               if path.find(literal) >= 0:
                   return True
       return False

    def command(self):
        """Returns the command required to execute the trigger script."""
        return ['/bin/sh', self._file]

def load_all(dirpath):
    if not os.path.isdir(dirpath):
        return []
    result = []
    for filename in os.listdir(dirpath):
        if not filename.endswith(Trigger.SUFFIX):
            continue
        filepath = os.path.join(dirpath, filename)
        p = Trigger(filepath)
        result.append(p)
    return result
        



# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2011 Red Hat, Inc.
#
#   fileutils.py - Helper functions for filesystem operations
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Author: Colin Walters <walters@verbum.org>

import os
import sys
import errno

def _accumulate_dirtree_contents_recurse(path, contents):
    names = os.listdir(path)
    for name in names:
        subpath = os.path.join(path, name)
        if os.path.isdir(subpath) and not os.path.islink(subpath):
            previous_len = len(contents)
            _accumulate_dirtree_contents_recurse(subpath, contents)
            new_len = len(contents)
            # Only add if the directory is empty, otherwise, its existence
            # is implicit.
            if previous_len == new_len:
                contents.append(subpath + os.sep)
        else:
            contents.append(subpath)

def accumulate_dirtree_contents(path):
    """Return a list of files and empty directories in the directory at PATH.  Each item
in the returned list is relative to the root path."""
    contents = []
    _accumulate_dirtree_contents_recurse(path, contents)
    if not path.endswith(os.sep):
        path = path + os.sep
    pathlen = len(path)
    for i,subpath in enumerate(contents):
        assert subpath.startswith(path)
        contents[i] = subpath[pathlen:]
    return contents

def remove_files_and_dirs(file_paths, allow_nonempty_dirs=False):
    """Given a list of file paths in any order, attempt to delete
them.  The main intelligence in this function is removing files
in a directory before removing the directory.

Returns a list, where each item is a 2-tuple:
(path, error_string or None)"""

    results = []

    for path in reversed(sorted(file_paths)):
        isdir = os.path.isdir(path) and not os.path.islink(path)
        try:
            if isdir:
                os.rmdir(path)
            else:
                os.unlink(path)
            results.append((path, True, ''))
        except OSError as e:
            if (isdir
                and allow_nonempty_dirs
                and len(os.listdir(path)) > 0):
                results.append((path, False, None))
            else:
                results.append((path, False, e.strerror))
    return results

def filter_files_by_prefix(config, file_paths):
    """Return the set of files in file_paths that are inside the prefix.

Turns relative paths into absolute paths, as appropriate."""
    canon_prefix = config.prefix
    if not canon_prefix.endswith(os.sep):
        canon_prefix = canon_prefix + os.sep
    result = []
    for path in file_paths:
        path = os.path.join(config.prefix, path) # doesn't add the prefix if path is already absolute
        if path == canon_prefix or (not path.startswith(canon_prefix)):
            continue
        result.append(path)
    return result

# Modified rename from http://selenic.com/repo/hg/file/tip/mercurial/windows.py
def _windows_rename(src, dst):
    '''atomically rename file src to dst, replacing dst if it exists'''
    try:
        os.rename(src, dst)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise
        # Windows does not allow to unlink open file.
        # Unlike in Mercurial, we don't try any workaround here.
        os.unlink(dst)
        os.rename(src, dst)

if os.name == 'nt':
    rename = _windows_rename
else:
    rename = os.rename

def ensure_unlinked(filename):
    try:
        os.unlink(filename)
    except OSError as e:
        if e.errno != os.errno.ENOENT:
            raise

def mkdir_with_parents(filename):
    try:
        os.makedirs(filename)
    except EnvironmentError as e:
        if e.errno != errno.EEXIST or not os.path.isdir(filename):
            raise

class SafeWriter(object):
    def __init__(self, filename):
        self.filename = filename
        self.tmpname = filename + '.tmp'
        self.fp = open(self.tmpname, 'w')

    def commit(self):
        self.fp.flush()
        if hasattr(os, 'fdatasync'):
            os.fdatasync(self.fp.fileno())
        else:
            os.fsync(self.fp.fileno())
        self.fp.close()

        rename(self.tmpname, self.filename)

    def abandon(self):
        self.fp.close()
        os.unlink(self.tmpname)



# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2001-2006  James Henstridge
# Copyright (C) 2007-2008  Frederic Peters
# Copyright (C) 2014 Canonical Limited
#
#   config.py: configuration file parser
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os
import os.path
import re
import sys
import traceback
import time
import types
import logging
import __builtin__

from jhbuild.environment import setup_env, setup_env_defaults, addpath
from jhbuild.errors import FatalError
from jhbuild.utils import sysid

if sys.platform.startswith('win'):
    # For munging paths for MSYS's benefit
    import jhbuild.utils.subprocess_win32

__all__ = [ 'Config' ]

_defaults_file = os.path.join(os.path.dirname(__file__), 'defaults.jhbuildrc')

_known_keys = [ 'moduleset', 'modules', 'skip', 'tags', 'prefix',
                'partial_build', 'checkoutroot', 'buildroot', 'top_builddir',
                'autogenargs', 'makeargs', 'nice_build', 'jobs',
                'installprog', 'repos', 'branches', 'noxvfb', 'xvfbargs',
                'builddir_pattern', 'module_autogenargs', 'module_makeargs',
                'module_ninjaargs', 'ninjaargs', 'interact', 'buildscript',
                'nonetwork', 'nobuild', 'alwaysautogen', 'noinstall',
                'makeclean', 'makedistclean', 'makecheck', 'module_makecheck',
                'system_libdirs', 'tinderbox_outputdir', 'sticky_date', 'tarballdir',
                'pretty_print', 'svn_program', 'makedist', 'makedistcheck',
                'nonotify', 'notrayicon', 'cvs_program', 'checkout_mode',
                'copy_dir', 'export_dir', 'module_checkout_mode', 'build_policy',
                'trycheckout', 'min_age', 'nopoison', 'module_nopoison',
                'forcecheck', 'makecheck_advisory', 'quiet_mode',
                'progress_bar', 'module_extra_env', 'jhbuildbot_master',
                'jhbuildbot_slavename', 'jhbuildbot_password',
                'jhbuildbot_svn_commits_box', 'jhbuildbot_slaves_dir',
                'jhbuildbot_dir', 'jhbuildbot_mastercfg',
                'use_local_modulesets', 'ignore_suggests', 'modulesets_dir',
                'mirror_policy', 'module_mirror_policy', 'dvcs_mirror_dir',
                'shallow_clone', 'build_targets', 'cmakeargs', 'module_cmakeargs',
                'mesonargs', 'module_mesonargs',
                'print_command_pattern', 'static_analyzer',
                'module_static_analyzer', 'static_analyzer_template',
                'static_analyzer_outputdir', 'check_sysdeps', 'system_prefix',
                'help_website', 'conditions', 'extra_prefixes',
                'disable_Werror', 'xdg_cache_home', 'exit_on_error'
              ]

env_prepends = {}
def prependpath(envvar, path):
    env_prepends.setdefault(envvar, []).append(path)

def parse_relative_time(s):
    m = re.match(r'(\d+) *([smhdw])', s.lower())
    if m:
        coeffs = {'s': 1, 'm': 60, 'h': 3600, 'd': 86400, 'w':7*86400}
        return float(m.group(1)) * coeffs[m.group(2)]
    else:
        raise ValueError

def modify_conditions(conditions, conditions_modifiers):
    for flag in conditions_modifiers:
        for mod in flag.split(','):
            if mod.startswith('+'):
                conditions.add(mod[1:])
            elif mod.startswith('-'):
                conditions.discard(mod[1:])
            else:
                raise FatalError(_("Invalid condition set modifier: '%s'.  Must start with '+' or '-'.") % mod)

class Config:
    _orig_environ = None

    def __init__(self, filename, conditions_modifiers):
        self._config = {
            '__file__': _defaults_file,
            'addpath':  addpath,
            'prependpath':  prependpath,
            'include': self.include,
            }

        if not self._orig_environ:
            self.__dict__['_orig_environ'] = os.environ.copy()
        os.environ['UNMANGLED_LD_LIBRARY_PATH'] = os.environ.get('LD_LIBRARY_PATH', '')
        os.environ['UNMANGLED_PATH'] = os.environ.get('PATH', '')

        env_prepends.clear()
        try:
            execfile(_defaults_file, self._config)
        except:
            traceback.print_exc()
            raise FatalError(_('could not load config defaults'))

        old_config = os.path.join(os.path.expanduser('~'), '.jhbuildrc')
        new_config = os.path.join \
                         (os.environ.get \
                             ('XDG_CONFIG_HOME',
                              os.path.join(os.path.expanduser('~'), '.config')),
                          'jhbuildrc')

        if filename:
            if not os.path.exists(filename):
                raise FatalError(_('could not load config file, %s is missing') % filename)
        else:
            if os.path.isfile(old_config) \
                and not os.path.islink(old_config) \
                and os.path.isfile(new_config) \
                and not os.path.islink(new_config):
                raise FatalError(_('The default location of the configuration '
                                   'file has changed. Please move %(old_path)s'
                                   ' to %(new_path)s.' \
                                   % {'old_path': old_config,
                                      'new_path': new_config}))
            if os.path.exists(new_config):
                filename = new_config
            elif os.path.exists(old_config):
                filename = old_config

        if filename:
            self._config['__file__'] = filename
            self.filename = filename
        else:
            self._config['__file__'] = new_config
            self.filename = new_config

        # we might need to redo this process on config reloads, so save these
        self.saved_conditions_modifiers = conditions_modifiers

        # We handle the conditions flags like so:
        #   - get the default set of conditions (determined by the OS)
        #   - modify it with the commandline arguments
        #   - load the config file so that it can make further modifications
        #   - modify it with the commandline arguments again
        #
        # We apply the commandline argument condition modifiers both before
        # and after parsing the configuration so that the jhbuildrc has a
        # chance to inspect the modified set of flags (and conditionally act
        # on it to set new autogenargs, for example) but also so that the
        # condition flags given on the commandline will ultimately override
        # those in jhbuildrc.
        self._config['conditions'] = sysid.get_default_conditions()
        modify_conditions(self._config['conditions'], conditions_modifiers)
        self.load(filename)
        modify_conditions(self.conditions, conditions_modifiers)

        self.create_directories()

        setup_env_defaults(self.system_libdirs)

        for prefix in reversed(self.extra_prefixes):
            setup_env(prefix)
        setup_env(self.prefix)

        self.apply_env_prepends()
        self.update_build_targets()

    def reload(self):
        os.environ = self._orig_environ.copy()
        self.__init__(filename=self._config.get('__file__'), conditions_modifiers=self.saved_conditions_modifiers)
        self.set_from_cmdline_options(options=None)

    def include(self, filename):
        '''Read configuration variables from a file.'''
        try:
            execfile(filename, self._config)
        except:
            traceback.print_exc()
            raise FatalError(_('Could not include config file (%s)') % filename)

    def load(self, filename=None):
        config = self._config
        if filename:
            try:
                execfile(filename, config)
            except Exception as e:
                if isinstance(e, FatalError):
                    # raise FatalErrors back, as it means an error in include()
                    # and it will print a traceback, and provide a meaningful
                    # message.
                    raise e
                traceback.print_exc()
                raise FatalError(_('could not load config file'))

        if not config.get('quiet_mode'):
            unknown_keys = []
            for k in config.keys():
                if k in _known_keys + ['cvsroots', 'svnroots', 'cflags']:
                    continue
                if k[0] == '_':
                    continue
                if type(config[k]) in (types.ModuleType, types.FunctionType, types.MethodType):
                    continue
                unknown_keys.append(k)
            if unknown_keys:
                logging.info(
                        _('unknown keys defined in configuration file: %s') % \
                        ', '.join(unknown_keys))

        # backward compatibility, from the days when jhbuild only
        # supported Gnome.org CVS.
        if config.get('cvsroot'):
            logging.warning(
                    _('the "%s" configuration variable is deprecated, '
                      'you should use "repos[\'gnome.org\']".') % 'cvsroot')
            config['repos'].update({'gnome.org': config['cvsroot']})
        if config.get('cvsroots'):
            logging.warning(
                    _('the "%s" configuration variable is deprecated, '
                      'you should use "repos".') % 'cvsroots')
            config['repos'].update(config['cvsroots'])
        if config.get('svnroots'):
            logging.warning(
                    _('the "%s" configuration variable is deprecated, '
                      'you should use "repos".') % 'svnroots')
            config['repos'].update(config['svnroots'])

        # environment variables
        if config.has_key('cflags') and config['cflags']:
            os.environ['CFLAGS'] = config['cflags']
        if config.get('installprog') and os.path.exists(config['installprog']):
            os.environ['INSTALL'] = config['installprog']

        for path_key in ('checkoutroot', 'buildroot', 'top_builddir',
                         'tinderbox_outputdir', 'tarballdir', 'copy_dir',
                         'jhbuildbot_slaves_dir', 'jhbuildbot_dir',
                         'jhbuildbot_mastercfg', 'modulesets_dir',
                         'dvcs_mirror_dir', 'static_analyzer_outputdir',
                         'prefix'):
            if config.get(path_key):
                config[path_key] = os.path.expanduser(config[path_key])

        # copy known config keys to attributes on the instance
        for name in _known_keys:
            setattr(self, name, config[name])

        # default tarballdir to checkoutroot
        if not self.tarballdir: self.tarballdir = self.checkoutroot

        # Ensure top_builddir is absolute
        if not os.path.isabs(self.top_builddir):
            self.top_builddir = os.path.join(self.prefix, self.top_builddir)

        # check possible checkout_mode values
        seen_copy_mode = (self.checkout_mode == 'copy')
        possible_checkout_modes = ('update', 'clobber', 'export', 'copy')
        if self.checkout_mode not in possible_checkout_modes:
            raise FatalError(_('invalid checkout mode'))
        for module, checkout_mode in self.module_checkout_mode.items():
            seen_copy_mode = seen_copy_mode or (checkout_mode == 'copy')
            if checkout_mode not in possible_checkout_modes:
                raise FatalError(_('invalid checkout mode (module: %s)') % module)
        if seen_copy_mode and not self.copy_dir:
            raise FatalError(_('copy mode requires copy_dir to be set'))

        if not os.path.exists(self.modulesets_dir):
            if self.use_local_modulesets:
                logging.warning(
                        _('modulesets directory (%s) not found, '
                          'disabling use_local_modulesets') % self.modulesets_dir)
                self.use_local_modulesets = False
            self.modulesets_dir = None

        if self.buildroot and not os.path.isabs(self.buildroot):
            raise FatalError(_('%s must be an absolute path') % 'buildroot')
        if not os.path.isabs(self.checkoutroot):
            raise FatalError(_('%s must be an absolute path') % 'checkoutroot')
        if not os.path.isabs(self.prefix):
            raise FatalError(_('%s must be an absolute path') % 'prefix')
        if not os.path.isabs(self.tarballdir):
            raise FatalError(_('%s must be an absolute path') % 'tarballdir')
        if (self.tinderbox_outputdir and
            not os.path.isabs(self.tinderbox_outputdir)):
            raise FatalError(_('%s must be an absolute path') %
                             'tinderbox_outputdir')

    def get_original_environment(self):
        return self._orig_environ

    def create_directories(self):
        if not os.path.exists(self.prefix):
            try:
                os.makedirs(self.prefix)
            except:
                raise FatalError(_('install prefix (%s) can not be created') % self.prefix)

        if not os.path.exists(self.top_builddir):
            try:
                os.makedirs(self.top_builddir)
            except OSError:
                raise FatalError(
                        _('working directory (%s) can not be created') % self.top_builddir)

        if os.path.exists(os.path.join(self.prefix, 'lib64', 'libglib-2.0.so')):
            raise FatalError(_("Your install prefix contains a 'lib64' directory, which is no longer "
                               "supported by jhbuild.  This is likely the result of a previous build with an "
                               "older version of jhbuild or of a broken package.  Please consider removing "
                               "your install and checkout directories and starting fresh."))

    def apply_env_prepends(self):
        ''' handle environment prepends ... '''
        for envvar in env_prepends.keys():
            for path in env_prepends[envvar]:
                addpath(envvar, path)

    def update_build_targets(self):
        # update build targets according to old flags
        if self.makecheck and not 'check' in self.build_targets:
            self.build_targets.insert(0, 'check')
        if self.makeclean and not 'clean' in self.build_targets:
            self.build_targets.insert(0, 'clean')
        if self.makedistclean and not 'distclean' in self.build_targets:
            self.build_targets.insert(0, 'distclean')
        if self.nobuild:
            # nobuild actually means "checkout"
            for phase in ('configure', 'build', 'check', 'clean', 'install'):
                if phase in self.build_targets:
                    self.build_targets.remove(phase)
            self.build_targets.append('checkout')
        if self.makedist and not 'dist' in self.build_targets:
            self.build_targets.append('dist')
        if self.makedistcheck and not 'distcheck' in self.build_targets:
            self.build_targets.append('distcheck')

    def set_from_cmdline_options(self, options=None):
        if options is None:
            options = self.cmdline_options
        else:
            self.cmdline_options = options
        if hasattr(options, 'autogen') and options.autogen:
            self.alwaysautogen = True
        if hasattr(options, 'check') and (
                options.check and not 'check' in self.build_targets):
            self.build_targets.insert(0, 'check')
        if hasattr(options, 'clean') and (
                options.clean and not 'clean' in self.build_targets):
            self.build_targets.insert(0, 'clean')
        if hasattr(options, 'distclean') and (
                options.distclean and not 'distclean' in self.build_targets):
            self.build_targets.insert(0, 'distclean')
        if hasattr(options, 'dist') and (
                options.dist and not 'dist' in self.build_targets):
            self.build_targets.append('dist')
        if hasattr(options, 'distcheck') and (
                options.distcheck and not 'distcheck' in self.build_targets):
            self.build_targets.append('distcheck')
        if hasattr(options, 'ignore_suggests') and options.ignore_suggests:
            self.ignore_suggests = True
        if hasattr(options, 'nonetwork') and options.nonetwork:
            self.nonetwork = True
        if hasattr(options, 'skip'):
            for item in options.skip:
                self.skip += item.split(',')
        if hasattr(options, 'tags'):
            for item in options.tags:
                self.tags += item.split(',')
        if hasattr(options, 'sticky_date') and options.sticky_date is not None:
                self.sticky_date = options.sticky_date
        if hasattr(options, 'xvfb') and options.noxvfb is not None:
                self.noxvfb = options.noxvfb
        if hasattr(options, 'trycheckout') and  options.trycheckout:
            self.trycheckout = True
        if hasattr(options, 'nopoison') and options.nopoison:
            self.nopoison = True
        if hasattr(options, 'quiet') and options.quiet:
            self.quiet_mode = True
        if hasattr(options, 'force_policy') and options.force_policy:
            self.build_policy = 'all'
        if hasattr(options, 'min_age') and options.min_age:
            try:
                self.min_age = time.time() - parse_relative_time(options.min_age)
            except ValueError:
                raise FatalError(_('Failed to parse \'min_age\' relative '
                                   'time'))
        if (hasattr(options, 'check_sysdeps') and
            options.check_sysdeps is not None):
            self.check_sysdeps = options.check_sysdeps

    def __setattr__(self, k, v):
        '''Override __setattr__ for additional checks on some options.'''
        if k == 'quiet_mode' and v:
            try:
                import curses
                logging.getLogger().setLevel(logging.ERROR)
            except ImportError:
                logging.warning(
                        _('quiet mode has been disabled because the Python curses module is missing.'))
                v = False

        self.__dict__[k] = v




# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2008 Igalia S.L., John Carr, Frederic Peters
#
#   steps.py: custom buildbot steps
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os, commands, re, StringIO

from buildbot.process import factory
from buildbot.process import buildstep
from buildbot import steps
from buildbot.status.builder import SUCCESS, WARNINGS, FAILURE, SKIPPED, EXCEPTION

class JHBuildSource(steps.source.Source):
    name = "jhbuild"

    def __init__ (self, moduleset=None, module=None, stage=None, **kwargs):
        steps.source.Source.__init__(self, **kwargs)
        self.moduleset = moduleset
        self.module = module
        self.name = module + ' update'
        self.description = [module + ' updating']
        self.descriptionDone = [module + ' updated']

    def computeSourceRevision(self, changes):
        if not changes:
            return None
        try:
            return max([int(c.revision) for c in changes])
        except ValueError:
            # in git, revisions are not integers, return last.
            return changes[-1].revision

    def startVC(self, branch, revision, patch):
        command = ['jhbuild']
        if (self.moduleset is not None):
            command += ['--moduleset='+self.moduleset]
        command += ['bot', '--step', 'updateone', self.module]
        properties = self.build.getProperties()
        kwargs = {}
        kwargs['workdir'] = './'
        #kwargs = properties.render(self.remote_kwargs)
        kwargs['command'] = properties.render(command)
        kwargs['env'] = {}
        kwargs['timeout'] = 60*60
        cmd = buildstep.RemoteShellCommand(**kwargs)
        self.startCommand(cmd)

class UnitTestsObserver(buildstep.LogLineObserver):

    def __init__(self):
        self.regroupfailed = []
        self.regrouppassed = []
        self.reunittest = []
        self.unittests = []
        buildstep.LogLineObserver.__init__(self)
        if len(self.regroupfailed) == 0:
            self.regroupfailed.append((re.compile('^(FAIL:) (.*)$'), 1))
        if len(self.regrouppassed) == 0:
            self.regrouppassed.append((re.compile('^(PASS:) (.*)$'), 1))
        if len(self.reunittest) == 0:
            self.reunittest.append((re.compile('^([^:]*):([^:]*):([^:]*):([^:]*):([^:]*):([^:]*).*$'), 4, 5))

    def outLineReceived(self, line):
        matched = False
        for r in self.regroupfailed:
            result = r[0].search(line)
            if result:
                self.step.failedTestsCount += 1
                self.step.testsResults.append((result.groups()[r[1]].strip(), False, self.unittests))
                self.unittests = []
                matched = True
        if not matched:
            for r in self.regrouppassed:
                result = r[0].search(line)
                if result:
                    self.step.passedTestsCount += 1
                    self.step.testsResults.append((result.groups()[r[1]].strip(), True, self.unittests))
                    self.unittests = []
                    matched = True
        if not matched:
            for r in self.reunittest:
                result = r[0].search(line)
                if result:
                    err_msg = result.groups()[r[2]].strip()
                    if err_msg == "Passed":
                        self.unittests.append((result.groups()[r[1]].strip(), True, err_msg))
                    else:
                        self.unittests.append((result.groups()[r[1]].strip(), False, err_msg))
                    matched = True

class JHBuildCommand(steps.shell.ShellCommand):
    name = "jhbuild_stage"
    haltOnFailure = 1
    description = ["jhbuild command"]
    descriptionDone = ["jhbuild"]
    command = None
    OFFprogressMetrics = ('output',)

    # things to track: number of files compiled, number of directories
    # traversed (assuming 'make' is being used)
    def __init__(self, stage=None,module=None, moduleset=None, **kwargs):
        assert module is not None
        kwargs['workdir'] = "./"
        #workdir = "./"
        command = ['jhbuild']
        if (moduleset is not None):
            command += ['--moduleset='+moduleset]
        command += ['bot', '--step', stage, module]
        self.name = module + ' ' + stage
        self.description = [module + ' ' + stage + ' (running)']
        self.descriptionDone = [module + ' ' + stage]
        steps.shell.ShellCommand.__init__(self, description=self.description,
                              descriptionDone=self.descriptionDone, command=command, **kwargs)

    def getText(self, cmd, results):
        text = self.describe(True)[:]
        return text


class JHBuildCheckCommand(JHBuildCommand):
    def __init__(self, **kwargs):
        JHBuildCommand.__init__(self, stage='check', **kwargs)
        self.failedTestsCount = 0
        self.passedTestsCount = 0
        self.testsResults = []
        self.addLogObserver('stdio', UnitTestsObserver())

    def evaluateCommand(self, cmd):
        if self.failedTestsCount > 0 or cmd.rc != 0:
            return WARNINGS
        else:
            return SUCCESS

    def getText(self, cmd, results):
        text = JHBuildCommand.getText(self, cmd, results)
        if self.failedTestsCount > 0 or self.passedTestsCount > 0:
            text.append('failed: %s' % self.failedTestsCount)
            text.append('passed: %s' % self.passedTestsCount)
        return text

    def createSummary(self, log):
        if self.failedTestsCount > 0 or self.passedTestsCount > 0:
            self.addHTMLLog('summary', self.createTestsSummary())

    def createTestsSummary (self):
        html = '<html>\n'
        html += '<head>\n'
        html += ' <title>Tests summary</title>\n'
        html += ' <link rel="stylesheet" type="text/css" href="/lgo.css">\n'
        html += ' <title>Tests summary</title>\n'
        html += '</head>\n'
        html += '<body>\n'

        # Show summary
        has_details = False
        html += '<ul id="tests-summary">\n'
        for t in self.testsResults:
            if t[1]:
                ttdclass = "success"
            else:
                ttdclass = "failure"
            if len(t[2]) > 0:
                html += '<li class="%s"><a href="#%s">%s</a></li>\n' % (ttdclass, t[0], t[0])
                has_details = True
            else:
                html += '<li class="%s">%s</a></li>\n' % (ttdclass, t[0])
        html += '</ul>\n'

        if has_details:
            # Show details
            html += '<dl id="test-details">\n'
            for t in self.testsResults:
                if len(t[2]) == 0:
                    continue
                if t[1]:
                    ttdclass = "success"
                else:
                    ttdclass = "failure"
                html += '<dt class="%s">%s</dt>\n' % (ttdclass, t[0])
                html += '<dd><ul>\n'

                for ut in t[2]:
                    if ut[1]:
                        uttdclass = "success"
                    else:
                        uttdclass = "failure"
                    html += '<li class="%s">' % uttdclass
                    html += ut[0]
                    if ut[1] == False:
                        html += ' ' + ut[2]
                    html += '</li>'
                html += '</ul></dd>\n'
            html += '</dl\n'

        html += '</body>\n'
        html += '</html>\n'

        return html


class JHBuildModulePathCommand(steps.shell.ShellCommand):
    name = "jhbuild_stage"
    haltOnFailure = 1
    description = ["jhbuild command"]
    descriptionDone = ["jhbuild"]
    command = None
    OFFprogressMetrics = ('output',)

    # things to track: number of files compiled, number of directories
    # traversed (assuming 'make' is being used)
    def __init__(self, module=None, moduleset=None, action='', actionName='', **kwargs):
        assert module is not None
        kwargs['workdir'] = "./"
        #workdir = "./"
        command = ["jhbuild"]
        if (moduleset is not None):
            command += ['--moduleset='+moduleset]
        actionParts = action.split(" ")
        command += ['run', '--in-builddir', module, '--']
        command += actionParts
        self.name = module + " " + actionName
        self.description = actionName + ' (running)'
        self.descriptionDone = [actionName]
        steps.shell.ShellCommand.__init__(self, description=self.description,
                                   descriptionDone=self.descriptionDone, command=command, **kwargs)

    def evaluateCommand(self, cmd):
        if self.haltOnFailure and cmd.rc != 0:
            return FAILURE
        elif cmd.rc != 0:
            return WARNINGS
        else:
            return SUCCESS	

    def createSummary(self, log):
        output = StringIO.StringIO(log.getText())
        warnings = []
        for line in output.readlines():
            if "warning:" in line:
                warnings.append(line)
            if "buildbot-url:" in line:
                arr = line.split()
                if len(arr) == 3:
                    self.addURL(arr[1], arr[2])
        if (len(warnings) > 0):
            self.addCompleteLog('warnings', "".join(warnings))

    def getText(self, cmd, results):
        text = self.describe(True)[:]
        return text




# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2001-2006  James Henstridge
# Copyright (C) 2007-2008  Frederic Peters
#
#   autotools.py: autotools module type definitions.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

__metaclass__ = type

import os
import stat
try:
    import hashlib
except ImportError:
    import md5 as hashlib

from jhbuild.errors import FatalError, BuildStateError, CommandError
from jhbuild.modtypes import \
     DownloadableModule, register_module_type, MakeModule
from jhbuild.versioncontrol.tarball import TarballBranch

__all__ = [ 'AutogenModule' ]

class AutogenModule(MakeModule, DownloadableModule):
    '''Base type for modules that are distributed with a Gnome style
    "autogen.sh" script and the GNU build tools.  Subclasses are
    responsible for downloading/updating the working copy.'''
    type = 'autogen'

    PHASE_CHECKOUT = DownloadableModule.PHASE_CHECKOUT
    PHASE_FORCE_CHECKOUT = DownloadableModule.PHASE_FORCE_CHECKOUT
    PHASE_CLEAN          = 'clean'
    PHASE_DISTCLEAN      = 'distclean'
    PHASE_CONFIGURE      = 'configure'
    PHASE_BUILD          = 'build'
    PHASE_CHECK          = 'check'
    PHASE_DIST           = 'dist'
    PHASE_INSTALL        = 'install'

    def __init__(self, name, branch=None,
                 autogenargs='', makeargs='',
                 makeinstallargs='',
                 supports_non_srcdir_builds=True,
                 skip_autogen=False,
                 skip_install_phase=False,
                 uninstall_before_install=False,
                 autogen_sh='autogen.sh',
                 makefile='Makefile',
                 autogen_template=None,
                 check_target=True,
                 supports_static_analyzer=True,
                 needs_gmake=True,
                 supports_unknown_configure_options=True):
        MakeModule.__init__(self, name, branch=branch, makeargs=makeargs,
                            makeinstallargs=makeinstallargs, makefile=makefile, needs_gmake=needs_gmake)
        self.autogenargs = autogenargs
        self.supports_non_srcdir_builds = supports_non_srcdir_builds
        self.force_non_srcdir_builds = False
        self.skip_autogen = skip_autogen
        self.skip_install_phase = skip_install_phase
        self.uninstall_before_install = uninstall_before_install
        self.autogen_sh = autogen_sh
        self.autogen_template = autogen_template
        self.check_target = check_target
        self.supports_install_destdir = True
        self.supports_static_analyzer = supports_static_analyzer
        self.supports_unknown_configure_options = supports_unknown_configure_options

    def get_srcdir(self, buildscript):
        return self.branch.srcdir

    def get_builddir(self, buildscript):
        builddir = self.get_srcdir(buildscript);
        if buildscript.config.buildroot and self.supports_non_srcdir_builds:
            d = buildscript.config.builddir_pattern % (
                self.branch.checkoutdir or self.branch.get_module_basename())
            builddir = os.path.join(buildscript.config.buildroot, d)
        if self.force_non_srcdir_builds and builddir == self.get_srcdir(buildscript):
            builddir = os.path.join(builddir, 'build')
        return builddir

    def _file_exists_and_is_newer_than(self, potential, other):
        try:
            other_stbuf = os.stat(other)
            potential_stbuf = os.stat(potential)
        except OSError as e:
            return False
        return potential_stbuf.st_mtime > other_stbuf.st_mtime

    def _get_configure_cmd(self, buildscript):
        '''returns a string of the command-line to configure the module.
        This method may modify self.autogen_sh, if 'autogen.sh' doesn't exist.
        FIXME: bad idea to modify self.autogen_sh in a getter.
        '''
        srcdir = self.get_srcdir(buildscript)
        if self.autogen_sh == 'autogen.sh' and \
        not os.path.exists(os.path.join(srcdir, self.autogen_sh)):
            # if there is no autogen.sh, automatically fallback to configure
            if os.path.exists(os.path.join(srcdir, 'configure')):
                self.autogen_sh = 'configure'

        if self.autogen_template:
            template = self.autogen_template
        else:
            template = ("%(srcdir)s/%(autogen-sh)s --prefix %(prefix)s %(autogenargs)s ")

        default_autogenargs = (self.config.autogenargs
            if self.supports_unknown_configure_options else '')
        autogenargs = self.autogenargs + ' ' + self.config.module_autogenargs.get(
                self.name, default_autogenargs)

        if self.config.disable_Werror and self.supports_unknown_configure_options:
            autogenargs = '--disable-Werror' + ' ' + autogenargs

        vars = {'prefix': os.path.splitdrive(buildscript.config.prefix)[1],
                'autogen-sh': self.autogen_sh,
                'autogenargs': autogenargs}

        if buildscript.config.buildroot and self.supports_non_srcdir_builds or \
           self.force_non_srcdir_builds:
            vars['srcdir'] = self.get_srcdir(buildscript)
        else:
            vars['srcdir'] = '.'

        cmd = self.static_analyzer_pre_cmd(buildscript) + template % vars

        if self.autogen_sh == 'autoreconf':
            cmd = cmd.replace('autoreconf', 'configure')
            cmd = cmd.replace('--enable-maintainer-mode', '')

        # if we are using configure as the autogen command, make sure
        # we don't pass --enable-maintainer-mode, since it breaks many
        # tarball builds.
        if self.autogen_sh == 'configure':
            cmd = cmd.replace('--enable-maintainer-mode', '')

        # If there is no --exec-prefix in the constructed autogen command, we
        # can safely assume it will be the same as {prefix} and substitute it
        # right now, so the printed command can be copy/pasted afterwards.
        # (GNOME #580272)
        if not '--exec-prefix' in template:
            cmd = cmd.replace('${exec_prefix}', vars['prefix'])

        self.configure_cmd = cmd
        return cmd

    def skip_configure(self, buildscript, last_phase):
        # skip if manually instructed to do so
        if self.skip_autogen is True:
            return True

        # don't skip this stage if we got here from one of the
        # following phases:
        if last_phase in [self.PHASE_FORCE_CHECKOUT,
                          self.PHASE_CLEAN,
                          self.PHASE_BUILD,
                          self.PHASE_INSTALL]:
            return False

        if self.skip_autogen == 'never':
            return False

        if buildscript.config.alwaysautogen:
            return False

        srcdir = self.get_srcdir(buildscript)
        configure_path = os.path.join(srcdir, 'configure')
        if not os.path.exists(configure_path):
            return False

        # if autogen.sh args has changed, re-run configure
        db_entry = buildscript.moduleset.packagedb.get(self.name)
        if db_entry:
            configure_hash = db_entry.metadata.get('configure-hash')
            if configure_hash:
                configure_cmd = self._get_configure_cmd(buildscript)
                if hashlib.md5(configure_cmd).hexdigest() != configure_hash:
                    return False
            else:
                # force one-time reconfigure if no configure-hash
                return False
        else:
            # always configure for the first build
            return False

        # We can't rely on the autotools maintainer-mode stuff because many
        # modules' autogen.sh script includes e.g. gtk-doc and/or intltool,
        # which also need to be rerun.
        # https://bugzilla.gnome.org/show_bug.cgi?id=660844
        if not isinstance(self.branch, TarballBranch):
            configsrc = None
            for name in ['configure.ac', 'configure.in']:
                path = os.path.join(srcdir, name)
                if os.path.exists(path):
                    configsrc = path
                    break
            if configsrc is not None:
                config_status = os.path.join(self.get_builddir(buildscript),
                                             'config.status')
                if self._file_exists_and_is_newer_than(config_status,
                                                       configsrc):
                    return True

        return False

    def do_configure(self, buildscript):
        builddir = self.get_builddir(buildscript)
        if not os.path.exists(builddir):
            os.makedirs(builddir)
        buildscript.set_action(_('Configuring'), self)

        cmd = self._get_configure_cmd(buildscript)

        srcdir = self.get_srcdir(buildscript)
        try:
            if not (os.stat(os.path.join(srcdir, self.autogen_sh))[stat.ST_MODE] & 0111):
                os.chmod(os.path.join(srcdir, self.autogen_sh), 0755)
        except:
            pass

        if self.autogen_sh == 'autoreconf':
            buildscript.execute(['autoreconf', '-fi'], cwd=srcdir)
            os.chmod(os.path.join(srcdir, 'configure'), 0755)

        buildscript.execute(cmd, cwd = builddir, extra_env = self.extra_env)
    do_configure.depends = [PHASE_CHECKOUT]
    do_configure.error_phases = [PHASE_FORCE_CHECKOUT,
            PHASE_CLEAN, PHASE_DISTCLEAN]

    def skip_clean(self, buildscript, last_phase):
        if 'distclean' in self.config.build_targets:
            return True
        builddir = self.get_builddir(buildscript)
        if not os.path.exists(builddir):
            return True
        if not os.path.exists(os.path.join(builddir, self.makefile)):
            return True
        return False

    def do_clean(self, buildscript):
        buildscript.set_action(_('Cleaning'), self)
        self.make(buildscript, 'clean')
    do_clean.depends = [PHASE_CONFIGURE]
    do_clean.error_phases = [PHASE_FORCE_CHECKOUT, PHASE_CONFIGURE]

    def do_build(self, buildscript):
        buildscript.set_action(_('Building'), self)
        self.make(buildscript, pre=self.static_analyzer_pre_cmd(buildscript))
    do_build.depends = [PHASE_CONFIGURE]
    do_build.error_phases = [PHASE_FORCE_CHECKOUT, PHASE_CONFIGURE,
            PHASE_CLEAN, PHASE_DISTCLEAN]

    def static_analyzer_pre_cmd(self, buildscript):
        if self.supports_static_analyzer and buildscript.config.module_static_analyzer.get(self.name, buildscript.config.static_analyzer):
            template = buildscript.config.static_analyzer_template + ' '
            outputdir = buildscript.config.static_analyzer_outputdir

            if not os.path.exists(outputdir):
                os.makedirs(outputdir)

            vars = {'outputdir': outputdir,
                    'module': self.name
                   }

            return template % vars
        return ''

    def skip_check(self, buildscript, last_phase):
        if not self.check_target:
            return True
        if self.name in buildscript.config.module_makecheck:
            return not buildscript.config.module_makecheck[self.name]
        if 'check' not in buildscript.config.build_targets:
            return True
        return False

    def do_check(self, buildscript):
        buildscript.set_action(_('Checking'), self)
        try:
            self.make(buildscript, 'check', pre=self.static_analyzer_pre_cmd(buildscript))
        except CommandError:
            if not buildscript.config.makecheck_advisory:
                raise
    do_check.depends = [PHASE_BUILD]
    do_check.error_phases = [PHASE_FORCE_CHECKOUT, PHASE_CONFIGURE]

    def do_dist(self, buildscript):
        buildscript.set_action(_('Creating tarball for'), self)
        self.make(buildscript, 'dist')
    do_dist.depends = [PHASE_CONFIGURE]
    do_dist.error_phases = [PHASE_FORCE_CHECKOUT, PHASE_CONFIGURE]

    def do_distcheck(self, buildscript):
        buildscript.set_action(_('Dist checking'), self)
        self.make(buildscript, 'distcheck')
    do_distcheck.depends = [PHASE_DIST]
    do_distcheck.error_phases = [PHASE_FORCE_CHECKOUT, PHASE_CONFIGURE]

    def do_install(self, buildscript):
        if self.uninstall_before_install:
            packagedb =  buildscript.moduleset.packagedb
            if packagedb.check(self.name):
                buildscript.set_action(_('Uninstalling old installed version'), self)
                packagedb.uninstall(self.name)

        buildscript.set_action(_('Installing'), self)
        destdir = self.prepare_installroot(buildscript)
        self.make(buildscript, self.makeinstallargs or 'install', makeargs='DESTDIR={}'.format(destdir))
        self.process_install(buildscript, self.get_revision())

    do_install.depends = [PHASE_BUILD]
    do_install.error_phases = [PHASE_CONFIGURE]

    def skip_install(self, buildscript, last_phase):
        return self.config.noinstall or self.skip_install_phase

    def skip_distclean(self, buildscript, last_phase):
        builddir = self.get_builddir(buildscript)
        if not os.path.exists(builddir):
            return True
        if not os.path.exists(os.path.join(builddir, self.makefile)) and \
           not hasattr(self.branch, 'delete_unknown_files'):
            return True

        return False

    def do_distclean(self, buildscript):
        buildscript.set_action(_('Distcleaning'), self)
        if hasattr(self.branch, 'delete_unknown_files'):
            self.branch.delete_unknown_files(buildscript)
        else:
            self.make(buildscript, 'distclean')
    do_distclean.depends = [PHASE_CHECKOUT]

    def xml_tag_and_attrs(self):
        return ('autotools',
                [('autogenargs', 'autogenargs', ''),
                 ('id', 'name', None),
                 ('makeargs', 'makeargs', ''),
                 ('makeinstallargs', 'makeinstallargs', ''),
                 ('supports-non-srcdir-builds',
                  'supports_non_srcdir_builds', True),
                 ('force-non-srcdir-builds',
                  'force_non_srcdir_builds', False),
                 ('supports-unknown-configure-options',
                  'supports_unknown_configure_options', True),
                 ('skip-autogen', 'skip_autogen', False),
                 ('skip-install', 'skip_install_phase', False),
                 ('uninstall-before-install', 'uninstall_before_install', False),
                 ('autogen-sh', 'autogen_sh', 'autogen.sh'),
                 ('makefile', 'makefile', 'Makefile'),
                 ('supports-static-analyzer', 'supports_static_analyzer', True),
                 ('autogen-template', 'autogen_template', None)])

def collect_args(instance, node, argtype):
    if node.hasAttribute(argtype):
        args = node.getAttribute(argtype)
    else:
        args = ''

    for child in node.childNodes:
        if child.nodeType == child.ELEMENT_NODE and child.nodeName == argtype:
            if not child.hasAttribute('value'):
                raise FatalError(_("<%s/> tag must contain value=''") % argtype)
            args += ' ' + child.getAttribute('value')

    return instance.eval_args(args)

def parse_autotools(node, config, uri, repositories, default_repo):
    instance = AutogenModule.parse_from_xml(node, config, uri, repositories, default_repo)

    # Allow base packages such as autoconf/automake/libtool/etc. to skip the
    # standard dependencies to prevent dependency cycles.
    if node.getAttribute('bootstrap') != 'true':
        instance.dependencies += ['automake', 'libtool', instance.get_makecmd(config)]

    instance.autogenargs = collect_args (instance, node, 'autogenargs')
    instance.makeargs = collect_args (instance, node, 'makeargs')
    instance.makeinstallargs = collect_args (instance, node, 'makeinstallargs')

    if node.hasAttribute('supports-non-srcdir-builds'):
        instance.supports_non_srcdir_builds = \
                (node.getAttribute('supports-non-srcdir-builds') != 'no')
    if node.hasAttribute('force-non-srcdir-builds'):
        instance.force_non_srcdir_builds = \
                (node.getAttribute('force-non-srcdir-builds') != 'no')
    if node.hasAttribute('supports-unknown-configure-options'):
        instance.supports_unknown_configure_options = \
                (node.getAttribute('supports-unknown-configure-options') != 'no')
    if node.hasAttribute('skip-autogen'):
        skip_autogen = node.getAttribute('skip-autogen')
        if skip_autogen == 'true':
            instance.skip_autogen = True
        elif skip_autogen == 'never':
            instance.skip_autogen = 'never'
    if node.hasAttribute('skip-install'):
        skip_install = node.getAttribute('skip-install')
        if skip_install.lower() in ('true', 'yes'):
            instance.skip_install_phase = True
        else:
            instance.skip_install_phase = False
    if node.hasAttribute('uninstall-before-install'):
        instance.uninstall_before_install = (node.getAttribute('uninstall-before-install') == 'true')

    if node.hasAttribute('check-target'):
        instance.check_target = (node.getAttribute('check-target') == 'true')
    if node.hasAttribute('supports-static-analyzer'):
        instance.supports_static_analyzer = (node.getAttribute('supports-static-analyzer') == 'true')

    from jhbuild.versioncontrol.tarball import TarballBranch
    if node.hasAttribute('autogen-sh'):
        autogen_sh = node.getAttribute('autogen-sh')
        if autogen_sh is not None:
            instance.autogen_sh = autogen_sh
        elif isinstance(instance.branch, TarballBranch):
            # in tarballs, force autogen-sh to be configure, unless autogen-sh is
            # already set
            instance.autogen_sh = 'configure'

    if node.hasAttribute('makefile'):
        instance.makefile = node.getAttribute('makefile')
    if node.hasAttribute('autogen-template'):
        instance.autogen_template = node.getAttribute('autogen-template')

    return instance
register_module_type('autotools', parse_autotools)




# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2001-2004  James Henstridge
#
#   autobuild.py: non-interactive build that generates a report
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from optparse import make_option

from jhbuild.errors import UsageError
from jhbuild.commands.base import Command, register_command
import jhbuild.frontends
import optparse

class cmd_autobuild(Command):
    doc = N_('Build modules non-interactively and upload results to JhAutobuild')

    name = 'autobuild'

    def __init__(self):
        Command.__init__(self, [
            make_option('-a', '--autogen',
                        action='store_true', dest='autogen', default=False,
                        help=_('always run autogen.sh')),
            make_option('-c', '--clean',
                        action='store_true', dest='clean', default=False,
                        help=_('run make clean before make')),
            make_option('--distcheck',
                        action='store_true', dest='distcheck', default=False,
                        help=_('run make distcheck after building')),
            make_option('-s', '--skip', metavar='MODULES',
                        action='append', dest='skip', default=[],
                        help=_('treat the given modules as up to date')),
            make_option('-t', '--start-at', metavar='MODULE',
                        action='store', dest='startat', default=None,
                        help=_('start building at the given module')),
            make_option('-r', '--report-url',
                        action='store', dest='reporturl', default=None,
                        help=_('jhautobuild report URL')),
            make_option('-v', '--verbose',
                        action='store_true', dest='verbose', default=False,
                        help=_('verbose mode')),
            ])

    def run(self, config, options, args, help=None):
        config.set_from_cmdline_options(options)
        config.buildscript = 'autobuild'

        config.autobuild_report_url = None
        config.verbose = False
        config.interact = False

        if options.reporturl is not None:
            config.autobuild_report_url = options.reporturl
        if options.verbose:
            config.verbose = True

        if not config.autobuild_report_url:
            raise UsageError(_('report URL for autobuild not specified'))
    
        module_set = jhbuild.moduleset.load(config)
        module_list = module_set.get_module_list(args or config.modules,
                                                 config.skip)
    
        # remove modules up to startat
        if options.startat:
            while module_list and module_list[0].name != options.startat:
                del module_list[0]
            if not module_list:
                raise FatalError(_('%s not in module list') % options.startat)
    
        build = jhbuild.frontends.get_buildscript(config, module_list, module_set=module_set)
        return build.build()

register_command(cmd_autobuild)



# jhbuild - a build script for GNOME 2.x
# Copyright (C) 2012  Frederic Peters
#
#   extdeps.py: report details on GNOME external dependencies
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from optparse import make_option
import re
import socket
import sys
import time

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

import jhbuild.moduleset
from jhbuild.commands import Command, register_command

HTML_AT_TOP = '''<html>
<head>
<title>%(title)s</title>
<style type="text/css">
body {
    font-family: sans-serif;
}
tbody th {
    background: #d3d7cf;
    text-align: left;
}
tbody td {
    text-align: center;
}

td { padding: 0 1em; }
td.rdeps { text-align: left; }

tr.zero { text-decoration: line-through; }
tr.far  { color: gray; }
tr.many { font-weight: bold; }

div#footer {
  margin-top: 2em;
  font-size: small;
}

</style>
</head>
<body>
<h1>%(title)s</h1>

'''


class cmd_extdeps(Command):
    doc = _('Report details on GNOME external dependencies')
    name = 'extdeps'

    def __init__(self):
        Command.__init__(self, [
            make_option('-o', '--output', metavar='FILE',
                    action='store', dest='output', default=None),
            make_option('--all-modules',
                        action='store_true', dest='list_all_modules', default=False),
            ])

    def run(self, config, options, args, help=None):
        if options.output:
            output = StringIO()
        else:
            output = sys.stdout

        config.partial_build = False
        self.module_set = jhbuild.moduleset.load(config)
        if options.list_all_modules:
            module_list = self.module_set.modules.values()
        else:
            module_list = self.module_set.get_module_list(args or config.modules, config.skip)

        if type(config.moduleset) is str:
            moduleset = [config.moduleset]
        else:
            moduleset = config.moduleset
        title = _('External deps for GNOME')
        for ms in moduleset:
            try:
                gnome_ver = re.findall('\d+\.\d+', ms)[0]
            except IndexError:
                continue
            title = _('External deps for GNOME %s') % gnome_ver
            break

        print >> output, HTML_AT_TOP % {'title': title}
        print >> output, '<table>'
        print >> output, '<tbody>'

        module_list.sort(lambda x,y: cmp(x.name.lower(), y.name.lower()))
        for mod in module_list:
            #if not mod.moduleset_name.startswith('gnome-suites-core-deps-base'):
            #    continue

            if not hasattr(mod.branch, 'version'):
                continue

            rdeps = self.compute_rdeps(mod)
            classes = []
            if len(rdeps) == 0:
                classes.append('zero')
            else:
                for rdep in rdeps:
                    rdep_mod = self.module_set.modules.get(rdep)
                    if not hasattr(rdep_mod.branch, 'version'):
                        break
                else:
                    # module who only has tarballs as dependency
                    classes.append('far')
                if len(rdeps) > 5:
                    classes.append('many')

            print >> output, '<tr class="%s">' % ' '.join(classes)
            print >> output, '<th>%s</th>' % mod.name
            version = mod.branch.version
            if mod.branch.patches:
                version = version + ' (%s)' % _('patched')
            print >> output, '<td class="version">%s</td>' % version
            print >> output, '<td class="url"><a href="%s">tarball</a></td>' % mod.branch.module
            if len(rdeps) > 5:
                rdeps = rdeps[:4] + [_('and %d others.')  % (len(rdeps)-4)]
            print >> output, '<td class="rdeps">%s</td>' % ', '.join(rdeps)
            print >> output, '</tr>'

        print >> output, '</tbody>'
        print >> output, '</table>'

        print >> output, '<div id="footer">'
        print >> output, 'Generated:', time.strftime('%Y-%m-%d %H:%M:%S %z')
        print >> output, 'on ', socket.getfqdn()
        print >> output, '</div>'

        print >> output, '</body>'
        print >> output, '</html>'

        if output != sys.stdout:
            file(options.output, 'w').write(output.getvalue())


    def compute_rdeps(self, module):
        rdeps = []
        for mod in self.module_set.modules.values():
            if mod.type == 'meta': continue
            if module.name in mod.dependencies:
                rdeps.append(mod.name)
        rdeps.sort(lambda x,y: cmp(x.lower(), y.lower()))
        return rdeps

register_command(cmd_extdeps)



# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2001-2006  James Henstridge
# Copyright (C) 2007  Gary Kramlich <grim@reaperworld.com>
#
#   mtn.py: some code to handle various Monotone operations
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

__all__ = []
__metaclass__ = type

import os

from jhbuild.errors import CommandError, FatalError
from jhbuild.utils.cmds import get_output
from jhbuild.versioncontrol import Repository, Branch, register_repo_type
from jhbuild.commands.sanitycheck import inpath

class MonotoneRepository(Repository):
    """A class representing a Monotone database."""

    init_xml_attrs = ['server', 'database', 'defbranch']

    def __init__(self, config, name, server, database, defbranch):
        Repository.__init__(self, config, name)

        self.server = config.repos.get(name, server)

        self.database = os.path.join(self.config.checkoutroot,
                                     config.repos.get(name, database))

        self.defbranch = config.repos.get(name, defbranch)

    branch_xml_attrs = ['branch', 'module', 'checkoutdir']

    def branch(self, name, branch=None, module=None, checkoutdir=None):
        if name in self.config.branches:
            module = self.config.branches[module]
            if not module:
                raise FatalError(_('branch for %(name)s has wrong override, check your %(filename)s') % \
                                   {'name'     : name,
                                    'filename' : self.config.filename})

        if not branch:
            branch = self.defbranch

        return MonotoneBranch(self, name, checkoutdir, branch, module)

class MonotoneBranch(Branch):
    """A class representing a Monotone branch."""

    def __init__(self, repository, name, checkoutdir, branch, module=None):
        Branch.__init__(self, repository, branch, checkoutdir)
        self.name = name
        self.branch = branch
        self.mtn_module = module

    def _codir(self):
        if self.checkoutdir:
            return os.path.join(self.checkoutroot, self.checkoutdir)
        elif self.mtn_module:
            return os.path.join(self.checkoutroot, self.branch)
        else:
            return os.path.join(self.checkoutroot, self.name)
    _codir = property(_codir)

    def srcdir(self):
        if self.mtn_module:
            return os.path.join(self._codir, self.mtn_module)
        else:
            return self._codir
    srcdir = property(srcdir)

    def branchname(self):
        return self.branch
    branchname = property(branchname)

    def _init(self, buildscript):
        """Initializes the monotone database"""

        buildscript.message(_('Initializing %s') % (self.repository.database))

        cmd = ['mtn', '-d', self.repository.database, 'db', 'init']
        buildscript.execute(cmd, 'mtn')

    def _pull(self, buildscript):
        """Pulls new revs into the database from the given server"""

        buildscript.message(_('Pulling branch %(branch)s from %(server)s') %
                            {'branch':self.branch, 'server':self.repository.server})

        cmd = ['mtn', '-d', self.repository.database, 'pull',
               self.repository.server, self.branch]

        buildscript.execute(cmd, 'mtn')

    def _check_for_conflict(self):
        """Checks 'mtn automate heads' for more than 1 head which would mean we have
          conflicts"""

        output = get_output(['mtn', 'automate', 'heads'],
                            cwd=self.srcdir)

        heads = len(output.splitlines())

        if heads > 1:
            raise CommandError(_('branch %(branch)s has %(num)d heads') %
                               {'branch':self.branch, 'num':heads})

    def _checkout(self, buildscript):
        """Checks out a branch from a repository."""

        buildscript.message(_('Checking out branch \'%(branch)s\' to directory \'%(dir)s\'') %
                            {'branch':self.branch, 'dir':self.srcdir})
        cmd = ['mtn', '-d', self.repository.database, 'co', '-b', self.branch,
               self._codir]
        buildscript.execute(cmd, 'mtn')

    def _update(self, buildscript):
        """Updates a monotone working directory."""

        buildscript.message(_('Updating working copy %s') % (self.srcdir))

        cmd = ['mtn', 'up']
        buildscript.execute(cmd, 'mtn', cwd=self._codir)

    def checkout(self, buildscript):
        # XXX: doesn't support alternative checkout modes
        if not inpath('mtn', os.environ['PATH'].split(os.pathsep)):
            raise CommandError(_('%s not found') % 'mtn')

        if not os.path.exists(self.repository.database):
            self._init(buildscript)

        self._pull(buildscript)

        if os.path.exists(self.srcdir):
            try:
                self._check_for_conflict()
            except CommandError:
                buildscript.execute(['mtn', 'heads'], 'mtn', cwd=self.srcdir)
                raise

            self._update(buildscript)
        else:
            self._checkout(buildscript)

    def tree_id(self):
        try:
            output = get_output(['mtn', 'automate', 'get_base_revision_id'],
                                cwd=self.srcdir)
        except CommandError:
            return None
        return output[0]


register_repo_type('mtn', MonotoneRepository)



# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2009 Frederic Peters
#
#   bot.py: pages with info on slaves
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import urllib, time

from twisted.web import html
from twisted.web.util import Redirect

from buildbot.status.web.base import HtmlResource

class JhBuildbotResource(HtmlResource):
    def __init__(self, slavename, req):
        HtmlResource.__init__(self)
        self.slavename = slavename
        botmaster = self.getStatus(req).botmaster
        self.slave = botmaster.slaves[slavename]

    def getTitle(self, request):
        return 'Buildbot: %s' % self.slave.slavename

    def body(self, req):
        data = ''
        data += '<dl>\n'
        data += '  <dt>Contact</dt>\n'
        data += '  <dd><ul>'
        if self.slave.contact_name:
            data += '<li>%s</li>\n' % self.slave.contact_name
        if self.slave.url:
            data += '<li><a href="%s">%s</a></li>\n' % (self.slave.url, self.slave.url)
        data += '  </ul></dd>'
        data += '  <dt>Running</dt>'
        data += '  <dd><ul>'
        if self.slave.distribution:
            data += '<li>%s</li>\n' % self.slave.distribution
        if self.slave.version:
            data += '<li>%s</li>\n' % self.slave.version
        if self.slave.architecture:
            data += '<li>%s</li>\n' % self.slave.architecture
        data += '  </ul></dd>'
        data += '</dl>'
        return data


class JhBuildbotsResource(HtmlResource):
    def getChild(self, path, req):
        if path == '':
            return Redirect('..')
        parent = req.site.buildbot_service
        if path in parent.slaves:
            return JhBuildbotResource(path, req)
        return HtmlResource.getChild(self, path, req)



# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2001-2006  James Henstridge
#
#   cvs.py: some code to handle various cvs operations
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

__all__ = [
    'CVSRepository',
    'login',
    'get_sticky_tag',
    ]
__metaclass__ = type

import sys
import os
try:
    import hashlib
except ImportError:
    import md5 as hashlib

import git

from jhbuild.errors import BuildStateError, CommandError
from jhbuild.versioncontrol import Repository, Branch, register_repo_type
from jhbuild.commands.sanitycheck import inpath
from jhbuild.utils.sxml import sxml


# table used to scramble passwords in ~/.cvspass files
_shifts = [
    0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15,
   16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31,
  114,120, 53, 79, 96,109, 72,108, 70, 64, 76, 67,116, 74, 68, 87,
  111, 52, 75,119, 49, 34, 82, 81, 95, 65,112, 86,118,110,122,105,
   41, 57, 83, 43, 46,102, 40, 89, 38,103, 45, 50, 42,123, 91, 35,
  125, 55, 54, 66,124,126, 59, 47, 92, 71,115, 78, 88,107,106, 56,
   36,121,117,104,101,100, 69, 73, 99, 63, 94, 93, 39, 37, 61, 48,
   58,113, 32, 90, 44, 98, 60, 51, 33, 97, 62, 77, 84, 80, 85,223,
  225,216,187,166,229,189,222,188,141,249,148,200,184,136,248,190,
  199,170,181,204,138,232,218,183,255,234,220,247,213,203,226,193,
  174,172,228,252,217,201,131,230,197,211,145,238,161,179,160,212,
  207,221,254,173,202,146,224,151,140,196,205,130,135,133,143,246,
  192,159,244,239,185,168,215,144,139,165,180,157,147,186,214,176,
  227,231,219,169,175,156,206,198,129,164,150,210,154,177,134,127,
  182,128,158,208,162,132,167,209,149,241,153,251,237,236,171,195,
  243,233,253,240,194,250,191,155,142,137,245,235,163,242,178,152
]

def scramble(password):
    return 'A' + ''.join([chr(_shifts[ord(ch)]) for ch in password])
def descramble(password):
    assert password[0] == 'A', 'unknown password format'
    return ''.join([chr(_shifts[ord(ch)]) for ch in password[1:]])

def _canonicalise_cvsroot(cvsroot):
    if not cvsroot.startswith(':pserver:'): return cvsroot
    parts = cvsroot.split(':')
    if parts[3].startswith('/'):
        parts[3] = '2401' + parts[3]
    return ':'.join(parts)

def login(cvsroot, password=None):
    if not cvsroot.startswith(':pserver:'): return
    cvsroot = _canonicalise_cvsroot(cvsroot)
    cvspass = os.path.join(os.environ['HOME'], '.cvspass')

    # cvs won't ask for this, so don't write it
    if password == '':
        return

    # check if the password has already been entered:
    try:
        fp = open(cvspass, 'r')
        for line in fp.readlines():
            parts = line.split()
            if not parts:
                continue
            if parts[0] == '/1':
                root = parts[1]
            else:
                root = _canonicalise_cvsroot(parts[0])
            if root == cvsroot:
                return
                break
    except IOError:
        pass
     # if we have a password, just write it directly to the .cvspass file
    if password is not None:
        fp = open(cvspass, 'a')
        fp.write('/1 %s %s\n' % (cvsroot, scramble(password)))
        fp.close()
    else:
        # call cvs login ..
        if os.system('cvs -d %s login' % cvsroot) != 0:
            sys.stderr.write(_('could not log into %s\n') % cvsroot)
            sys.exit(1)

def check_sticky_tag(filename):
    dirname = os.path.dirname(filename)
    basename = os.path.basename(filename)
    entries_file = os.path.join(dirname, 'CVS', 'Entries')
    fp = open(entries_file, 'r')
    line = fp.readline()
    while line:
        parts = line.strip().split('/')
        if parts[1] == basename:
            # parts[5] is the tag for this file
            if parts[5] == '':
                return None
            else:
                return parts[5][1:]
        line = fp.readline()
    raise RuntimeError(_('%s is not managed by CVS') % filename)

def check_root(dirname):
    root_file = os.path.join(dirname, 'CVS', 'Root')
    return open(root_file, 'r').read().strip()

def _process_directory(directory, prefix, write):
    if not (os.path.isdir(directory) and
            os.path.isdir(os.path.join(directory, 'CVS'))):
        return
    
    fp = open(os.path.join(directory, 'CVS', 'Root'), 'rb')
    root = fp.read().strip()
    fp.close()
    fp = open(os.path.join(directory, 'CVS', 'Repository'), 'rb')
    repository = fp.read().strip()
    fp.close()

    write('===\n')
    write('Directory: %s\n' % prefix)
    write('Root: %s\n' % root)
    write('Repository: %s\n' % repository)
    write('\n')

    fp = open(os.path.join(directory, 'CVS', 'Entries'), 'rb')
    subdirs = []
    filenames = []
    for line in fp:
        parts = line.strip().split('/')
        if parts[0] == 'D' and len(parts) >= 2:
            subdirs.append(parts[1])
        if parts[0] == '' and len(parts) >= 3:
            filenames.append((parts[1], parts[2]))
    fp.close()
    filenames.sort()
    for name, rev in filenames:
        if prefix:
            name = '%s/%s' % (prefix, name)
        write('%s %s\n' % (name, rev))
    write('\n')
    subdirs.sort()
    for name in subdirs:
        if prefix:
            name_prefix = '%s/%s' % (prefix, name)
        else:
            name_prefix = name
        _process_directory(os.path.join(directory, name), name_prefix, write)



class CVSRepository(Repository):
    """A class used to work with a CVS repository"""

    init_xml_attrs = ['cvsroot', 'password']

    def __init__(self, config, name, cvsroot, password=None):
        Repository.__init__(self, config, name)
        # has the repository path been overridden?
        if self.name in config.repos:
            self.cvsroot = config.repos[name]
        else:
            self.cvsroot = cvsroot
            login(cvsroot, password)
        self.cvs_program = config.cvs_program

    branch_xml_attrs = ['module', 'checkoutdir', 'revision',
                        'update-new-dirs', 'override-checkoutdir']

    def branch(self, name, module=None, checkoutdir=None, revision=None,
               update_new_dirs='yes', override_checkoutdir='yes'):
        if module is None:
            module = name
        # allow remapping of branch for module:
        revision = self.config.branches.get(name, revision)
        if self.cvs_program == 'git-cvsimport':
            return git.GitCvsBranch(repository=self,
                                    module=module,
                                    checkoutdir=checkoutdir,
                                    revision=revision)
        else:
            return CVSBranch(repository=self,
                         module=module,
                         checkoutdir=checkoutdir,
                         revision=revision,
                         update_new_dirs=update_new_dirs != 'no',
                         override_checkoutdir=override_checkoutdir != 'no')

    def to_sxml(self):
        return [sxml.repository(type='cvs', name=self.name, cvsroot=self.cvsroot)]

    def get_sysdeps(self):
        return ['cvs']


class CVSBranch(Branch):
    """A class representing a CVS branch inside a CVS repository"""

    def __init__(self, repository, module, checkoutdir, revision,
                 update_new_dirs, override_checkoutdir):
        Branch.__init__(self, repository, module, checkoutdir)
        self.revision = revision
        self.update_new_dirs = update_new_dirs
        self.override_checkoutdir = override_checkoutdir

    def srcdir(self):
        if self.checkoutdir:
            return os.path.join(self.checkoutroot, self.checkoutdir)
        else:
            return os.path.join(self.checkoutroot, self.module)
    srcdir = property(srcdir)

    def branchname(self):
        return self.revision
    branchname = property(branchname)

    def _export(self, buildscript):
        cmd = ['cvs', '-z3', '-q', '-d', self.repository.cvsroot,
               'export']
        if self.revision:
            cmd.extend(['-r', self.revision])
        else:
            cmd.extend(['-r', 'HEAD'])
        if self.config.sticky_date:
            cmd.extend(['-D', self.config.sticky_date])
        if self.checkoutdir and self.override_checkoutdir:
            cmd.extend(['-d', self.checkoutdir])
        cmd.append(self.module)
        buildscript.execute(cmd, 'cvs', cwd=self.config.checkoutroot)

    def _checkout(self, buildscript, copydir=None):
        cmd = ['cvs', '-z3', '-q', '-d', self.repository.cvsroot,
               'checkout', '-P']
        if self.revision:
            cmd.extend(['-r', self.revision])
        if self.config.sticky_date:
            cmd.extend(['-D', self.config.sticky_date])
        if not (self.revision or self.config.sticky_date):
            cmd.append('-A')
        if self.checkoutdir and self.override_checkoutdir:
            cmd.extend(['-d', self.checkoutdir])
        cmd.append(self.module)
        if copydir:
            buildscript.execute(cmd, 'cvs', cwd=copydir)
        else:
            buildscript.execute(cmd, 'cvs', cwd=self.config.checkoutroot)

    def _update(self, buildscript, copydir=None):
        # sanity check the existing working tree:
        if copydir:
            outputdir = os.path.join(copydir, os.path.basename(self.srcdir))
        else:
            outputdir = self.srcdir
        try:
            wc_root = check_root(outputdir)
        except IOError:
            raise BuildStateError(_('"%s" does not appear to be a CVS working copy')
                                  % os.path.abspath(outputdir))
        if wc_root != self.repository.cvsroot:
            raise BuildStateError(_('working copy points at the wrong repository (expected %(root1)s but got %(root2)s). ') 
                                  % {'root1':self.repository.cvsroot, 'root2':wc_root} +
                                  _('Consider using the changecvsroot.py script to fix this.'))

        # update the working tree
        cmd = ['cvs', '-z3', '-q', '-d', self.repository.cvsroot,
               'update', '-P']
        if self.update_new_dirs:
            cmd.append('-d')
        if self.revision:
            cmd.extend(['-r', self.revision])
        if self.config.sticky_date:
            cmd.extend(['-D', self.config.sticky_date])
        if not (self.revision or self.config.sticky_date):
            cmd.append('-A')
        cmd.append('.')
        buildscript.execute(cmd, 'cvs', cwd=outputdir)

    def checkout(self, buildscript):
        if not inpath('cvs', os.environ['PATH'].split(os.pathsep)):
            raise CommandError(_('%s not found') % 'cvs')
        Branch.checkout(self, buildscript)

    def tree_id(self):
        if not os.path.exists(self.srcdir):
            return None
        md5sum = hashlib.md5()
        _process_directory(self.srcdir, '', md5sum.update)
        return 'jhbuild-cvs-treeid:%s' % md5sum.hexdigest()

    def to_sxml(self):
        # FIXME: fix the current revision
        return [sxml.branch(repo=self.repository.name,
                            module=self.module)]


register_repo_type('cvs', CVSRepository)



# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2001-2006  James Henstridge
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# extras not found in old versions of Python

import sys
import __builtin__

# Windows lacks all sorts of subprocess features that we need to kludge around
if sys.platform.startswith('win'):
    from jhbuild.utils import subprocess_win32
    sys.modules['subprocess'] = subprocess_win32



# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2001-2006  James Henstridge
#
#   testmodule.py: testmodule type definitions.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

__metaclass__ = type

import os, time, signal, sys, subprocess, random, tempfile

try:
    import hashlib
except ImportError:
    import md5 as hashlib

from jhbuild.errors import FatalError, CommandError, BuildStateError
from jhbuild.modtypes import \
     Package, DownloadableModule, register_module_type
from jhbuild.modtypes.autotools import AutogenModule

import xml.dom.minidom

__all__ = ['TestModule']
__test_types__ = ['ldtp' , 'dogtail']

class TestModule(Package, DownloadableModule):
    type = 'test'
    
    PHASE_CHECKOUT = DownloadableModule.PHASE_CHECKOUT
    PHASE_FORCE_CHECKOUT = DownloadableModule.PHASE_FORCE_CHECKOUT
    PHASE_TEST           = 'test'
    
    def __init__(self, name, branch=None, test_type=None, tested_pkgs=[]):
        Package.__init__(self, name, branch=branch)
        self.test_type    = test_type
        self.tested_pkgs  = tested_pkgs

        ### modify environ for tests to be working
        if os.environ.has_key('LDTP_DEBUG'):
            del os.environ['LDTP_DEBUG'] # get rid of verbose LDTP output
        if not os.environ.has_key('GNOME_ACCESSIBILITY') or os.environ['GNOME_ACCESSIBILITY'] != 1:
            os.environ['GNOME_ACCESSIBILITY'] = '1'

    def get_srcdir(self, buildscript):
        return self.branch.srcdir

    def _get_display(self):
        # get free display
        servernum = 99
        while True:
            if not os.path.exists('/tmp/.X%s-lock' % servernum):
                break
            servernum += 1
        return str(servernum)

    def _set_xauth(self, servernum):
        # create auth file
        paths = os.environ.get('PATH').split(':')
        flag = False
        for path in paths:
            if os.path.exists(os.path.join(path, 'xauth')):
                flag = True
                break
        tmpdir = tempfile.gettempdir()
        if not flag or os.path.exists(os.path.join(tmpdir,'jhbuild.%s' % os.getpid())):
            return ''

        try:
            os.mkdir(os.path.join(tmpdir,'jhbuild.%s' % os.getpid()))
            new_xauth = os.path.join(tmpdir, 'jhbuild.%s' % os.getpid(),'Xauthority')
            open(new_xauth, 'w').close()
            hexdigest = hashlib.md5(str(random.random())).hexdigest()
            os.system('xauth -f "%s" add ":%s" "." "%s"' % (
                        new_xauth, servernum, hexdigest))
        except OSError:
            return ''
        return new_xauth
    
    def do_test(self, buildscript):
        buildscript.set_action('Testing', self)
        if not buildscript.config.noxvfb:
            # start Xvfb
            old_display = os.environ.get('DISPLAY')
            old_xauth   = os.environ.get('XAUTHORITY')
            xvfb_pid = self._start_xvfb(buildscript.config.xvfbargs)
            if xvfb_pid == -1:
                raise BuildStateError('Unable to start Xvfb')

        # either do_ldtp_test or do_dogtail_test
        method = getattr(self, 'do_' + self.test_type + '_test')
        try:
            method(buildscript)
        finally:
            if not buildscript.config.noxvfb:
                # kill Xvfb if it has been started
                self._stop_xvfb(xvfb_pid)
                if old_display:
                    os.environ['DISPLAY'] = old_display
                else:
                    os.unsetenv('DISPLAY')
                if old_xauth:
                    os.environ['XAUTHORITY'] = old_xauth
                else:
                    os.unsetenv('XAUTHORITY')
    do_test.depends = [PHASE_CHECKOUT]

    def get_ldtp_log_file(self, filename):
        # <ldtp>
        # |
        # -- <logfile>filename</logfile>
        run_file = xml.dom.minidom.parse(filename)
        try:
            return run_file.getElementsByTagName('ldtp')[0].getElementsByTagName(
                    'logfile')[0].childNodes[0].data
        except IndexError:
            return None

    def _get_ldtp_info(self, node):
        infos = []
        errors = []
        warnings = []
        causes = []
        for info in node.getElementsByTagName('info'):
            for child in info.childNodes:
                infos.append(child.data)
        for cause in node.getElementsByTagName('cause'):
            for child in cause.childNodes:
                causes.append(child.data)
        for error in node.getElementsByTagName('error'):
            for child in error.childNodes:
                errors.append(child.data)
        for warning in node.getElementsByTagName('warning'):
            for child in warning.childNodes:
                warnings.append(child.data)
        return infos, errors, causes, warnings
        
    def _check_ldtp_log_file(self, logfile):
        log_file = xml.dom.minidom.parse(logfile)
        ldtp_node = log_file.getElementsByTagName('ldtp')[0]

        groups = []
        for group in ldtp_node.getElementsByTagName('group'):
            scr = []
            for script in group.getElementsByTagName('script'):
                tests = {}
                for test in script.getElementsByTagName('test'):
                    test_name = test.getAttribute('name')
                    
                    pass_status = test.getElementsByTagName('pass')[0].childNodes[0].data
                    infos, errors, causes, warnings = self._get_ldtp_info(test)
                    tests[test_name] = {
                        'pass': pass_status,
                        'info': infos,
                        'error': errors,
                        'cause': causes,
                        'warning': warnings
                    }

                infos, errors, causes, warnings = self._get_ldtp_info(script)
                scr.append({
                        'tests': tests,
                        'info': infos,
                        'error': errors,
                        'cause': causes,
                        'warning': warnings})

            groupstatus = group.getElementsByTagName('groupstatus')[0].childNodes[0].data
            groups.append({'script': scr, 'groupstatus': groupstatus})
        return groups

    def check_groups(self, groups):
        group_num = 1
        flag = False
        status = ''
        for group in groups:
            status += 'In Group #%s (%s)\n' % (group_num, group['groupstatus'])
            for script in group['script']:
                for test in script['tests'].keys():
                    status += 'Test \'%s\' ' % test
                    if script['tests'][test]['pass'] == '0': # failed
                        status += 'failed\n\tErrors'
                        for error in script['tests'][test]['error']:
                            status += ', '
                            status += error

                        status += '\n\tCauses'
                        for cause in script['tests'][test]['cause']:
                            status += ', '
                            status += cause

                        status += '\n\tWarnings'
                        for warning in script['tests'][test]['warning']:
                            status += ', '
                            status += warning

                        status += '\n\tInfos'
                        for info in script['tests'][test]['info']:
                            status += ', '
                            status += info                        
                    else:
                        status += 'passed'
                    status += '\n'

            group_num += 1
            if self._check_ldtp_group_status(group['groupstatus']):
                flag = True
        return flag, status
    
    def _check_ldtp_group_status(self, status):
        status = status.split()
        if status[0] != status[-1]:
            return True
        return False

    def _start_xvfb(self, xvfbargs):
        new_display = self._get_display()
        new_xauth   = self._set_xauth(new_display)
        if new_xauth == '':
            return -1

        os.environ['DISPLAY'] = ':' + new_display
        os.environ['XAUTHORITY'] = new_xauth
        try:
            xvfb = subprocess.Popen(
                    ['Xvfb',':'+new_display] + xvfbargs.split(), shell=False)
            self.screennum = new_display
            self.xauth = new_xauth
        except OSError:
            return -1
        
        time.sleep(2) #allow Xvfb to start
        if xvfb.poll() != None:
            return -1
        return xvfb.pid

    def _stop_xvfb(self, xvfb_pid):
        os.kill(xvfb_pid, signal.SIGINT)
        os.system('xauth remove ":%s"' % self.screennum)
        os.system('rm -r %s' % os.path.split(self.xauth)[0])
            
    def _start_ldtp(self):
        try:
            ldtp = subprocess.Popen('ldtp', shell=False)
        except OSError:
            return -1
        time.sleep(1)
        if ldtp.poll() != None:
            return -1
        return ldtp.pid
    
    def do_ldtp_test(self, buildscript):
        src_dir = self.get_srcdir(buildscript)
        old_debug = os.getenv('LDTP_DEBUG')
        if old_debug != None:
            del os.environ['LDTP_DEBUG']

        ldtp_pid = self._start_ldtp()
        if ldtp_pid == -1:
            raise BuildStateError('Unable to start ldtp server')

        try:
            if buildscript.config.noxvfb:
                buildscript.execute('ldtprunner run.xml', cwd=src_dir)
            else:
                buildscript.execute('ldtprunner run.xml', cwd=src_dir,
                        extra_env={'DISPLAY': ':%s' % self.screennum})
        except CommandError as e:
            os.kill(ldtp_pid, signal.SIGINT)
            if e.returncode == 32512:        # ldtprunner not installed
                raise BuildStateError('ldtprunner not available')
            raise BuildStateError('error %s during test' % e.returncode)
        os.kill(ldtp_pid, signal.SIGINT)
        
        if old_debug != None:
            os.environ['LDTP_DEBUG'] = old_debug
        
        log_file = self.get_ldtp_log_file(os.path.join (src_dir,'run.xml'))
        if not log_file:
            raise BuildStateError('missing log file')
        try:
            groups = self._check_ldtp_log_file(log_file)
            flag, status = self.check_groups(groups)
            if flag:
                raise BuildStateError(status)
        except:
            raise BuildStateError('malformed log file')

    def do_dogtail_test(self, buildscript):
        src_dir = self.get_srcdir(buildscript)
        test_cases = []
        failed = False
        all_files = os.listdir(src_dir)
        for file in all_files:
            if file[-3:] == '.py':
                test_cases.append(file)

        status = ''
        if buildscript.config.noxvfb:
            extra_env = {}
        else:
            extra_env = {'DISPLAY': ':%s' % self.screennum}

        for test_case in test_cases:
            try:
                buildscript.execute('python %s' % test_case,
                        cwd=src_dir, extra_env=extra_env)
            except CommandError as e:
                if e.returncode != 0:
                    raise BuildStateError('%s failed' % test_case)

    def xml_tag_and_attrs(self):
        return 'testmodule', [('id', 'name', None),
                              ('type', 'test_type', None)]

def get_tested_packages(node):
    tested_pkgs = []
    for tested_module in node.getElementsByTagName('testedmodules'):
        for mod in tested_module.getElementsByTagName('tested'):
            tested_pkgs.append(mod.getAttribute('package'))
    return tested_pkgs

def parse_testmodule(node, config, uri, repositories, default_repo):
    instance = TestModule.parse_from_xml(node, config, uri, repositories, default_repo)

    test_type = node.getAttribute('type')
    if test_type not in __test_types__:
        # FIXME: create an error here
        pass
    instance.test_type = test_type

    instance.tested_pkgs = get_tested_packages(node)
    
    return instance
                                   
register_module_type('testmodule', parse_testmodule)



# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2011  Colin Walters <walters@verbum.org>
#
#   systeminstall.py - Use system-specific means to acquire dependencies
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os
import sys 
import logging
import shlex
import subprocess
import pipes
import imp
import time
from StringIO import StringIO

import cmds

def get_installed_pkgconfigs(config):
    """Returns a dictionary mapping pkg-config names to their current versions on the system."""
    pkgversions = {}
    try:
        proc = subprocess.Popen(['pkg-config', '--list-all'], stdout=subprocess.PIPE, close_fds=True)
        stdout = proc.communicate()[0]
        proc.wait()
        pkgs = []
        for line in StringIO(stdout):
            pkg, rest = line.split(None, 1)
            pkgs.append(pkg)

        # see if we can get the versions "the easy way"
        try:
            stdout = subprocess.check_output(['pkg-config', '--modversion'] + pkgs)
            versions = stdout.splitlines()
            if len(versions) == len(pkgs):
                return dict(zip(pkgs, versions))
        except (subprocess.CalledProcessError, OSError):
            pass

        # We have to rather inefficiently repeatedly fork to work around
        # broken pkg-config installations - if any package has a missing
        # dependency pkg-config will fail entirely.
        for pkg in pkgs:
            args = ['pkg-config', '--modversion']
            args.append(pkg)
            proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True)
            stdout = proc.communicate()[0]
            proc.wait()
            pkgversions[pkg] = stdout.strip()
    except (subprocess.CalledProcessError, OSError): # pkg-config not installed
        pass
    return pkgversions

def get_uninstalled_pkgconfigs(uninstalled):
    uninstalled_pkgconfigs = []
    for module_name, dep_type, value in uninstalled:
        if dep_type == 'pkgconfig':
            uninstalled_pkgconfigs.append((module_name, value))
    return uninstalled_pkgconfigs

def get_uninstalled_binaries(uninstalled):
    uninstalled_binaries = []
    for module_name, dep_type, value in uninstalled:
        if dep_type.lower() == 'path':
            uninstalled_binaries.append((module_name, value))
    return uninstalled_binaries

def get_uninstalled_c_includes(uninstalled):
    uninstalled_c_includes = []
    for module_name, dep_type, value in uninstalled:
        if dep_type.lower() == 'c_include':
            uninstalled_c_includes.append((module_name, value))
    return uninstalled_c_includes

# This function returns uninstalled binaries and C includes as filenames.
# Backends should use either this OR the functions above.
def get_uninstalled_filenames(uninstalled):
    uninstalled_filenames = []
    for module_name, dep_type, value in uninstalled:
        if dep_type.lower() == 'path':
            uninstalled_filenames.append((module_name, os.path.join('/usr/bin', value)))
        elif dep_type.lower() == 'c_include':
            uninstalled_filenames.append((module_name, os.path.join('/usr/include', value)))
    return uninstalled_filenames

def systemdependencies_met(module_name, sysdeps, config):
    '''Returns True of the system dependencies are met for module_name'''
    def get_c_include_search_paths(config):
        '''returns a list of C include paths (-I) from the environment and the
        user's config'''
        def extract_path_from_cflags(args):
            '''extract the C include paths (-I) from a list of arguments (args)
            Returns a list of paths'''
            itr = iter(args.split())
            paths = []
            if os.name == 'nt':
                # shlex.split doesn't handle sep '\' on Windows
                import string
                shell_split = string.split
            else:
                shell_split = shlex.split
            try:
                while True:
                    arg = itr.next()
                    if arg.strip() in ['-I', '-isystem']:
                        # extract paths handling quotes and multiple paths
                        paths += shell_split(itr.next())[0].split(os.pathsep)
                    elif arg.startswith('-I'):
                        paths += shell_split(arg[2:])[0].split(os.pathsep)
            except StopIteration:
                pass
            return paths
        try:
            multiarch = subprocess.check_output(['gcc', '-print-multiarch']).strip()
        except:
            multiarch = None
        # search /usr/include and its multiarch subdir (if any) by default
        paths = [ os.path.join(os.sep, 'usr', 'include')]
        if multiarch:
            paths += [ os.path.join(paths[0], multiarch) ]
        paths += extract_path_from_cflags(os.environ.get('CPPFLAGS', ''))
        # check include paths incorrectly configured in CFLAGS, CXXFLAGS
        paths += extract_path_from_cflags(os.environ.get('CFLAGS', ''))
        paths += extract_path_from_cflags(os.environ.get('CXXFLAGS', ''))
        # check include paths incorrectly configured in makeargs
        paths += extract_path_from_cflags(config.makeargs)
        paths += extract_path_from_cflags(config.module_autogenargs.get
                                             (module_name, ''))
        paths += extract_path_from_cflags(config.module_makeargs.get
                                             (module_name, ''))
        paths += os.environ.get('C_INCLUDE_PATH', '').split(':')
        paths += os.environ.get('CPLUS_INCLUDE_PATH', '').split(':')
        paths = list(set(paths)) # remove duplicates
        return paths

    c_include_search_paths = None
    for dep_type, value, altdeps in sysdeps:
        dep_met = True
        if dep_type.lower() == 'path':
            if os.path.split(value)[0]:
                if not os.path.isfile(value) and not os.access(value, os.X_OK):
                    dep_met = False
            else:
                pathdirs = set(os.environ.get('PATH', '').split(os.pathsep))
                pathdirs.update(['/sbin', '/usr/sbin'])
                for path in pathdirs:
                    filename = os.path.join(path, value)
                    if os.path.isfile(filename) and os.access(filename, os.X_OK):
                        break
                else:
                    dep_met = False
        elif dep_type.lower() == 'c_include':
            if c_include_search_paths is None:
                c_include_search_paths = get_c_include_search_paths(config)
            found = False
            for path in c_include_search_paths:
                filename = os.path.join(path, value)
                if os.path.isfile(filename):
                    found = True
                    break
            if not found:
                dep_met = False

        elif dep_type == 'python2':
            try:
                imp.find_module(value)
            except:
                dep_met = False

        elif dep_type == 'xml':
            xml_catalog = '/etc/xml/catalog'

            if not os.path.exists(xml_catalog):
                for d in os.environ['XDG_DATA_DIRS'].split(':'):
                    xml_catalog = os.path.join(d, 'xml', 'catalog')
                    if os.path.exists(xml_catalog):
                        break

            try:
                # no xmlcatalog installed will (correctly) fail the check
                subprocess.check_output(['xmlcatalog', xml_catalog, value])

            except:
                dep_met = False

        # check alternative dependencies
        if not dep_met and altdeps:
            for altdep in altdeps:
                if systemdependencies_met(module_name, [ altdep ], config):
                    dep_met = True
                    break

        if not dep_met:
            return False

    return True

class SystemInstall(object):
    def __init__(self):
        if cmds.has_command('pkexec'):
            self._root_command_prefix_args = ['pkexec']
        elif cmds.has_command('sudo'):
            self._root_command_prefix_args = ['sudo']
        else:
            raise SystemExit(_('No suitable root privilege command found; you should install "sudo" or "pkexec" (or the system package that provides it)'))

    def install(self, uninstalled, assume_yes):
        """Takes a list of pkg-config identifiers and uses a system-specific method to install them."""
        raise NotImplementedError()

    @classmethod
    def find_best(cls):
        global _classes
        for possible_cls in _classes:
            if possible_cls.detect():
                return possible_cls()

# PackageKit dbus interface contains bitfield constants which
# aren't introspectable
PK_PROVIDES_ANY = 1
PK_FILTER_ENUM_NOT_INSTALLED = 1 << 3
PK_FILTER_ENUM_NEWEST = 1 << 16
PK_FILTER_ENUM_ARCH = 1 << 18
PK_TRANSACTION_FLAG_ENUM_ONLY_TRUSTED = 1 << 1

class PKSystemInstall(SystemInstall):
    def __init__(self):
        SystemInstall.__init__(self)
        self._loop = None
        self._sysbus = None
        self._pkdbus = None

    def _on_pk_message(self, msgtype, msg):
        logging.info(_('PackageKit: %s' % (msg,)))

    def _on_pk_error(self, msgtype, msg):
        logging.error(_('PackageKit: %s' % (msg,)))

    def _get_new_transaction(self):
        if self._loop is None:
            try:
                import glib
            except:
                try:
                    from gi.repository import GLib as glib
                except:
                    raise SystemExit(_('Error: python-gobject package not found.'))
            self._loop = glib.MainLoop()
        if self._sysbus is None:
            try:
                import dbus.glib
            except:
                raise SystemExit(_('Error: dbus-python package not found.'))
            import dbus
            self._dbus = dbus
            self._sysbus = dbus.SystemBus()
        if self._pkdbus is None:
            self._pkdbus = dbus.Interface(self._sysbus.get_object('org.freedesktop.PackageKit',
                                                                  '/org/freedesktop/PackageKit'),
                                          'org.freedesktop.PackageKit')
            properties = dbus.Interface(self._pkdbus, 'org.freedesktop.DBus.Properties')
        txn_path = self._pkdbus.CreateTransaction()
        txn = self._sysbus.get_object('org.freedesktop.PackageKit', txn_path)
        txn_tx = self._dbus.Interface(txn, 'org.freedesktop.PackageKit.Transaction')
        txn.connect_to_signal('Message', self._on_pk_message)
        txn.connect_to_signal('ErrorCode', self._on_pk_error)
        txn.connect_to_signal('Destroy', lambda *args: self._loop.quit())
        return txn_tx, txn

    def install(self, uninstalled, assume_yes):
        logging.info(_('Computing packages to install. This might be slow. Please wait.'))
        pk_package_ids = set()
        uninstalled_pkgconfigs = get_uninstalled_pkgconfigs(uninstalled)
        if uninstalled_pkgconfigs:
            txn_tx, txn = self._get_new_transaction()
            txn.connect_to_signal('Package', lambda info, pkid, summary: pk_package_ids.add(pkid))
            txn_tx.WhatProvides(PK_FILTER_ENUM_ARCH | PK_FILTER_ENUM_NEWEST |
                                PK_FILTER_ENUM_NOT_INSTALLED,
                                ['pkgconfig(%s)' % pkg for modname, pkg in
                                 uninstalled_pkgconfigs])
            self._loop.run()
            del txn, txn_tx

        uninstalled_filenames = get_uninstalled_filenames(uninstalled)
        if uninstalled_filenames:
            txn_tx, txn = self._get_new_transaction()
            txn.connect_to_signal('Package', lambda info, pkid, summary: pk_package_ids.add(pkid))
            txn_tx.SearchFiles(PK_FILTER_ENUM_ARCH | PK_FILTER_ENUM_NEWEST |
                               PK_FILTER_ENUM_NOT_INSTALLED,
                               [pkg for modname, pkg in
                                uninstalled_filenames])
            self._loop.run()
            del txn, txn_tx

        # On Fedora 17 a file can be in two packages: the normal package and
        # an older compat- package. Don't install compat- packages.
        pk_package_ids = [pkg for pkg in pk_package_ids
                          if not pkg.startswith('compat-')]

        if len(pk_package_ids) == 0:
            logging.info(_('Nothing available to install'))
            return

        logging.info(_('Installing:\n  %s' % ('\n  '.join(pk_package_ids, ))))
        logging.info(_('This might take a very long time. Do not turn off your computer. You can run `pkmon\' to monitor progress.'))

        txn_tx, txn = self._get_new_transaction()
        txn_tx.InstallPackages(PK_TRANSACTION_FLAG_ENUM_ONLY_TRUSTED, pk_package_ids)
        self._loop.run()

        logging.info(_('Complete!'))

    @classmethod
    def detect(cls):
        return cmds.has_command('pkcon')

class PacmanSystemInstall(SystemInstall):
    def __init__(self):
        SystemInstall.__init__(self)
        self._pacman_install_args = ['pacman', '-S', '--asdeps', '--needed', '--quiet', '--noconfirm']

    def _maybe_update_pkgfile(self):
        if not cmds.has_command('pkgfile'):
            logging.info(_('pkgfile not found, automatically installing'))
            if subprocess.call(self._root_command_prefix_args + self._pacman_install_args + ['pkgfile',]):
                logging.error(_('Failed to install pkgfile'))
                raise SystemExit

        # Update the pkgfile cache if it is older than 1 day.
        cacheexists = bool(os.listdir('/var/cache/pkgfile'))
        if not cacheexists or os.stat('/var/cache/pkgfile').st_mtime < time.time() - 86400:
            logging.info(_('pkgfile cache is old or doesn\'t exist, automatically updating'))
            result = subprocess.call(self._root_command_prefix_args + ['pkgfile', '--update'])
            if result and not cacheexists:
                logging.error(_('Failed to create pkgfile cache'))
                raise SystemExit
            elif result:
                logging.warning(_('Failed to update pkgfile cache'))
            else:
                logging.info(_('Successfully updated pkgfile cache'))

    def install(self, uninstalled, assume_yes):
        uninstalled_pkgconfigs = get_uninstalled_pkgconfigs(uninstalled)
        uninstalled_filenames = get_uninstalled_filenames(uninstalled)
        logging.info(_('Using pacman to install packages.  Please wait.'))
        package_names = set()

        if not uninstalled_filenames and not uninstalled_pkgconfigs:
            logging.info(_('Nothing to install'))
            return

        self._maybe_update_pkgfile()

        for name, pkgconfig in uninstalled_pkgconfigs:
            # Just throw the pkgconfigs in the normal file list
            uninstalled_filenames.append((None, '/usr/lib/pkgconfig/%s.pc' %pkgconfig))

        for name, filename in uninstalled_filenames:
            try:
                result = subprocess.check_output(['pkgfile', '--raw', filename])
                if result:
                    package_names.add(result.split('\n')[0])
            except subprocess.CalledProcessError:
                logging.warning(_('Provider for "%s" was not found, ignoring' %(name if name else filename)))

        if not package_names:
            logging.info(_('Nothing to install'))
            return

        logging.info('Installing:\n  %s' %('\n  '.join(package_names)))
        if subprocess.call(self._root_command_prefix_args + self._pacman_install_args + list(package_names)):
            logging.error(_('Install failed'))
        else:
            logging.info(_('Completed!'))

    @classmethod
    def detect(cls):
        if cmds.has_command('pacman'):
            return True
        return False

class AptSystemInstall(SystemInstall):
    def __init__(self):
        SystemInstall.__init__(self)

    def _get_package_for(self, filename, exact_match):
        if exact_match:
            proc = subprocess.Popen(['apt-file', '--fixed-string', 'search', filename],
                                    stdout=subprocess.PIPE, close_fds=True)
        else:
            proc = subprocess.Popen(['apt-file', 'search', filename],
                                    stdout=subprocess.PIPE, close_fds=True)
        stdout = proc.communicate()[0]
        if proc.returncode != 0:
            return None
        for line in StringIO(stdout):
            parts = line.split(':', 1)
            if len(parts) != 2:
                continue
            name = parts[0]
            path = parts[1]
            # Ignore copies of the pkg-config files that are not from the
            # libraries.
            if '/lsb3' in path:
                continue
            if '/emscripten' in path:
                continue

            # otherwise for now, just take the first match
            return name

    def _try_append_native_package(self, modname, filename, native_packages, exact_match):
        native_pkg = self._get_package_for(filename, exact_match)
        if native_pkg:
            native_packages.append(native_pkg)
            return True
        return False

    def _append_native_package_or_warn(self, modname, filename, native_packages, exact_match):
        if not self._try_append_native_package(modname, filename, native_packages, exact_match):
            logging.info(_('No native package found for %(id)s '
                           '(%(filename)s)') % {'id'       : modname,
                                                'filename' : filename})

    def _install_packages(self, native_packages, assume_yes):
        logging.info(_('Installing: %(pkgs)s') % {'pkgs': ' '.join(native_packages)})
        apt_cmd_line = ['apt-get']
        if assume_yes is True:
            apt_cmd_line += ['--assume-yes']
        apt_cmd_line += ['install']
        args = self._root_command_prefix_args + apt_cmd_line
        args.extend(native_packages)
        subprocess.check_call(args)

    def install(self, uninstalled, assume_yes):
        logging.info(_('Using apt-file to search for providers; this may be extremely slow. Please wait. Patience!'))
        native_packages = []

        pkgconfigs = [(modname, '/%s.pc' % pkg) for modname, pkg in
                      get_uninstalled_pkgconfigs(uninstalled)]
        for modname, filename in pkgconfigs:
            self._append_native_package_or_warn(modname, filename, native_packages, False)

        binaries = [(modname, '/usr/bin/%s' % pkg) for modname, pkg in
                    get_uninstalled_binaries(uninstalled)]
        for modname, filename in binaries:
            self._append_native_package_or_warn(modname, filename, native_packages, True)

        # Get multiarch include directory, e.g. /usr/include/x86_64-linux-gnu
        multiarch = None
        try:
            multiarch = subprocess.check_output(['gcc', '-print-multiarch']).strip()
        except:
            # Really need GCC to continue. Yes, this is fragile.
            self._install_packages(['gcc'])
            multiarch = subprocess.check_output(['gcc', '-print-multiarch']).strip()

        c_includes = get_uninstalled_c_includes(uninstalled)
        for modname, filename in c_includes:
            # Try multiarch first, so we print the non-multiarch location on failure.
            if (multiarch == None or
                not self._try_append_native_package(modname, '/usr/include/%s/%s' % (multiarch, filename), native_packages, True)):
                self._append_native_package_or_warn(modname, '/usr/include/%s' % filename, native_packages, True)

        if native_packages:
            self._install_packages(native_packages, assume_yes=assume_yes)
        else:
            logging.info(_('Nothing to install'))

    @classmethod
    def detect(cls):
        return cmds.has_command('apt-file')

# Ordered from best to worst
_classes = [AptSystemInstall, PacmanSystemInstall, PKSystemInstall]

if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    installer = SystemInstall.find_best()
    print "Using %r" % (installer, )
    installer.install(sys.argv[1:])



# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2008  Igalia S.L., John Carr, Frederic Peters
#
#   scheduler.py: jhbuild jobs scheduler 
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import time

from twisted.application import service, internet
from twisted.python import log

from twisted.internet import reactor
from buildbot.scheduler import Periodic, BaseUpstreamScheduler, Scheduler, Nightly
from buildbot.sourcestamp import SourceStamp
from buildbot import buildset, util

def SerialScheduler(name, project, builderNames, periodicBuildTimer=60*60*12,
        upstream=None, branch=None):
    if not upstream:
        return StartSerial(name, project, builderNames, periodicBuildTimer, branch)
    return Serial(name, project, upstream, builderNames, branch)

def NightlySerialScheduler(name, project, builderNames,
        minute=0, hour='*', dayOfMonth='*', month='*', dayOfWeek='*',
        upstream=None, branch=None):
    if not upstream:
        return NightlyStartSerial(name, project, builderNames,
                minute, hour, dayOfMonth, month, dayOfWeek, branch)
    return Serial(name, project, upstream, builderNames, branch)


class OnCommitScheduler(Scheduler):
    '''
    Scheduler that will build a module when a change notification
    (on svn-commits-list) is received.
    '''
    def __init__(self, name, project, builderNames, properties={}):
        Scheduler.__init__(self, name, branch=None, treeStableTimer=180,
                builderNames=builderNames, properties=properties)
        self.project = project
        self.importantChanges = []
        self.unimportantChanges = []
        self.nextBuildTime = None
        self.timer = None

    def changeIsImportant(self, change):
        if not change.files:
            # strange, better be on the safe side
            return True
        non_po_files = [x for x in change.files if not '/po/' in x]
        if non_po_files:
            return True
        # changes are limited to translations, it is unlikely it would break
        # the build, mark them as unimportant.
        return False

    def addChange(self, change):
        if change.project != self.project:
            return
        if change.branch != self.branch:
            return
        log.msg('adding a change')
        if self.changeIsImportant(change):
            self.addImportantChange(change)
        else:
            self.addUnimportantChange(change)

class StartSerial(Periodic):

    def __init__(self, name, project, builderNames, periodicBuildTimer,
                 branch=None):
        Periodic.__init__(self,name,builderNames,periodicBuildTimer,branch)
        self.project = project
        self.finishedWatchers = []

    def subscribeToFinishedBuilds(self, watcher):
        self.finishedWatchers.append(watcher)

    def unsubscribeToFinishedBuilds(self, watcher):
        self.finishedWatchers.remove(watcher)

    def buildSetFinished(self, bss):
        if not self.running:
            return
        ss = bss.getSourceStamp()
        for w in self.finishedWatchers:
            w(ss)
        Periodic.buildSetFinished(self,bss)

class Serial(BaseUpstreamScheduler):
    """This scheduler runs some set of builds that should be run
    after the 'upstream' scheduler has completed (successfully or not)."""
    compare_attrs = ('name', 'upstream', 'builders', 'branch')

    def __init__(self, name, project, upstream, builderNames, branch):
        BaseUpstreamScheduler.__init__(self, name)
        self.project = project
        self.upstream = upstream
        self.branch = branch
        self.builderNames = builderNames
        self.finishedWatchers = []

    def subscribeToFinishedBuilds(self, watcher):
        self.finishedWatchers.append(watcher)

    def unsubscribeToFinishedBuilds(self, watcher):
        self.finishedWatchers.remove(watcher)

    def buildSetFinished(self, bss):
        if not self.running:
            return
        ss = bss.getSourceStamp()
        for w in self.finishedWatchers:
            w(ss)
        BaseUpstreamScheduler.buildSetFinished(self,bss)

    def listBuilderNames(self):
        return self.builderNames

    def getPendingBuildTimes(self):
        # report the upstream's value
        return self.upstream.getPendingBuildTimes()

    def startService(self):
        service.MultiService.startService(self)
        self.upstream.subscribeToFinishedBuilds(self.upstreamBuilt)

    def stopService(self):
        d = service.MultiService.stopService(self)
        self.upstream.unsubscribeToFinishedBuilds(self.upstreamBuilt)
        return d

    def upstreamBuilt(self, ss):
        bs = buildset.BuildSet(self.builderNames, SourceStamp(branch=self.branch))
        self.submitBuildSet(bs)


class NightlyStartSerial(Nightly):
    def __init__(self, name, project, builderNames,
                 minute=0, hour='*', dayOfMonth='*', month='*', dayOfWeek='*',
                 branch=None):
        Nightly.__init__(self, name, builderNames, minute, hour, dayOfMonth,
                         month, dayOfWeek, branch)
        self.project = project
        self.finishedWatchers = []

    def subscribeToFinishedBuilds(self, watcher):
        self.finishedWatchers.append(watcher)

    def unsubscribeToFinishedBuilds(self, watcher):
        self.finishedWatchers.remove(watcher)

    def buildSetFinished(self, bss):
        if not self.running:
            return
        ss = bss.getSourceStamp()
        for w in self.finishedWatchers:
            w(ss)
        Nightly.buildSetFinished(self,bss)




# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2001-2006  James Henstridge
#
#   __init__.py: a package holding the various build frontends
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import sys

def get_buildscript(config, module_list=None, module_set=None):
    modname = 'jhbuild.frontends.%s' % config.buildscript
    __import__(modname)
    BuildScript = sys.modules[modname].BUILD_SCRIPT
    return BuildScript(config, module_list, module_set=module_set)



# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2001-2004  James Henstridge
#
#   checkbranches.py: check GNOME module sets for missing branches definition
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from optparse import make_option

import jhbuild.moduleset
from jhbuild.commands import Command, register_command
from jhbuild.utils.cmds import get_output
from jhbuild.errors import CommandError

class cmd_checkbranches(Command):
    doc = N_('Check modules in GNOME Git repository have the correct branch definition')
    name = 'checkbranches'
    
    def __init__(self):
        Command.__init__(self, [
            make_option('-b', '--branch', metavar = 'BRANCH',
                    action = 'store', dest = 'branch', default = None)])

    def run(self, config, options, args, help=None):
        if options.branch:
            branch = options.branch
        else:
            if type(config.moduleset) is list:
                branch = config.moduleset[0].replace('.', '-')
            else:
                branch = config.moduleset.replace('.', '-')
            for prefix in ('gnome-suites-core-deps', 'gnome-suites-core',
                           'gnome-suites-', 'gnome-apps-'):
                branch = branch.replace(prefix, 'gnome-')

        module_set = jhbuild.moduleset.load(config)
        module_list = module_set.get_module_list(args or config.modules)
        for mod in module_list:
            if mod.type in ('meta', 'tarball'):
                continue
            if not mod.branch or not mod.branch.repository.__class__.__name__ == 'GitRepository':
                continue
            if not 'git.gnome.org' in mod.branch.repository.href:
                continue
            if mod.branch.branch:
                # there is already a branch defined
                continue

            try:
                if get_output(['git', 'ls-remote',
                        'https://git.gnome.org/browse/%s' % mod.name,
                        'refs/heads/%s' % branch]):
                    uprint(_('%(module)s is missing branch definition for %(branch)s') % {'module': mod.name, 'branch': branch})
            except CommandError:
                pass


register_command(cmd_checkbranches)



# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2011 Colin Walters <walters@verbum.org>
#
#   make.py: Run build for cwd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import logging
import os
import pipes
import sys
import time
from optparse import make_option

import jhbuild.moduleset
import jhbuild.frontends
from jhbuild.errors import FatalError
from jhbuild.commands import Command, register_command

class cmd_make(Command):
    doc = N_('Compile and install the module for the current directory')

    name = 'make'
    usage_args = N_('[ options ... ]')

    def __init__(self):
        Command.__init__(self, [
            make_option('-a', '--autogen',
                        action='store_true', dest='autogen', default=False,
                        help=_('always run autogen.sh')),
            make_option('-c', '--clean',
                        action='store_true', dest='clean', default=False,
                        help=_('run make clean before make')),
            make_option('--check',
                        action='store_true', dest='check', default=False,
                        help=_('run make check after building')),
            make_option('-q', '--quiet',
                        action='store_true', dest='quiet', default=False,
                        help=_('quiet (no output)')),
            ])

    def run(self, config, options, args, help=None):
        # Grab the cwd before anything changes it
        cwd = self.get_cwd()

        # Explicitly don't touch the network for this
        options.nonetwork = True
        options.force_policy = True
        config.set_from_cmdline_options(options)

        makeargs = config.makeargs
        for arg in args:
            # if uninstalling, skip install.
            if arg == 'uninstall' or arg.startswith('uninstall-'):
                config.noinstall = True
            # pipes.quote (and really, trying to safely quote shell arguments) is
            # broken, but executing commands as strings is pervasive throughout
            # jhbuild...this is a hack that will probably live until someone just
            # replaces jhbuild entirely.
            makeargs = '%s %s' % (makeargs, pipes.quote(arg))
        config.makeargs = makeargs

        module_set = jhbuild.moduleset.load(config)

        if not cwd.startswith(config.checkoutroot):
            logging.error(_('The current directory is not in the checkout root %r') % (config.checkoutroot, ))
            return False

        cwd = cwd[len(config.checkoutroot):]
        cwd = cwd.lstrip(os.sep)
        modname, _slash, _rest = cwd.partition(os.sep)

        try:
            module = module_set.get_module(modname, ignore_case=True)
        except KeyError as e:
            default_repo = jhbuild.moduleset.get_default_repo()
            if not default_repo:
                logging.error(_('No module matching current directory %r in the moduleset') % (modname, ))
                return False

            # Try meson first, then autotools
            if os.path.exists(os.path.join(self.get_cwd(), 'meson.build')):
                from jhbuild.modtypes.meson import MesonModule
                module = MesonModule(modname, default_repo.branch(modname))
            else:
                from jhbuild.modtypes.autotools import AutogenModule
                module = AutogenModule(modname, default_repo.branch(modname))

            module.config = config
            logging.info(_('module "%(modname)s" does not exist, created automatically using repository "%(reponame)s"') % \
                         {'modname': modname, 'reponame': default_repo.name})

        build = jhbuild.frontends.get_buildscript(config, [module], module_set=module_set)
        return build.build()

register_command(cmd_make)




# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2001-2006  James Henstridge
#
#   meson.py: meson module type definitions.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

__metaclass__ = type

import os
import shutil

from jhbuild.errors import BuildStateError, CommandError
from jhbuild.modtypes import \
     Package, DownloadableModule, register_module_type, NinjaModule
from jhbuild.modtypes.autotools import collect_args
from jhbuild.commands.sanitycheck import inpath
from jhbuild.utils import fileutils

__all__ = [ 'MesonModule' ]

class MesonModule(NinjaModule, DownloadableModule):
    """Base type for modules that use Meson build system."""
    type = 'meson'

    PHASE_CHECKOUT = DownloadableModule.PHASE_CHECKOUT
    PHASE_FORCE_CHECKOUT = DownloadableModule.PHASE_FORCE_CHECKOUT
    PHASE_CLEAN = 'clean'
    PHASE_CONFIGURE = 'configure'
    PHASE_BUILD = 'build'
    PHASE_CHECK = 'check'
    PHASE_DIST = 'dist'
    PHASE_INSTALL = 'install'

    def __init__(self, name, branch=None,
                 mesonargs='', ninjaargs='',
                 skip_install_phase=False):
        NinjaModule.__init__(self, name, branch=branch, ninjaargs=ninjaargs)
        self.mesonargs = mesonargs
        self.supports_non_srcdir_builds = True
        self.skip_install_phase = skip_install_phase
        self.force_non_srcdir_builds = True
        self.supports_install_destdir = True

    def eval_args(self, args):
        args = Package.eval_args(self, args)
        args = args.replace('${libsuffix}', '')
        return args

    def get_libdir(self):
        return 'lib'

    def get_srcdir(self, buildscript):
        return self.branch.srcdir

    def get_builddir(self, buildscript):
        builddir = self.get_srcdir(buildscript)
        if buildscript.config.buildroot and self.supports_non_srcdir_builds:
            d = buildscript.config.builddir_pattern % (
                self.branch.checkoutdir or self.branch.get_module_basename())
            builddir = os.path.join(buildscript.config.buildroot, d)
        if self.force_non_srcdir_builds and builddir == self.get_srcdir(buildscript):
            builddir = os.path.join(builddir, 'build')
        return builddir

    def get_mesonargs(self):
        args = '%s %s' % (self.mesonargs,
                          self.config.module_mesonargs.get(
                              self.name, self.config.mesonargs))
        return self.eval_args(args)

    def do_configure(self, buildscript):
        buildscript.set_action(_('Configuring'), self)
        srcdir = self.get_srcdir(buildscript)
        builddir = self.get_builddir(buildscript)
        # meson does not allow configuring if the builddir already exists,
        # so we'll need to get rid of it and start over to configure again.
        if os.path.exists(builddir):
            shutil.rmtree(builddir)
        os.makedirs(builddir)
        prefix = os.path.expanduser(buildscript.config.prefix)
        if not inpath('meson', os.environ['PATH'].split(os.pathsep)):
            raise CommandError(_('%s not found') % 'meson')
        baseargs = '--prefix %s --libdir %s' % (prefix, self.get_libdir())
        mesonargs = self.get_mesonargs()
        cmd = 'meson %s %s %s' % (baseargs, mesonargs, srcdir)
        buildscript.execute(cmd, cwd=builddir, extra_env=self.extra_env)
    do_configure.depends = [PHASE_CHECKOUT]
    do_configure.error_phases = [PHASE_FORCE_CHECKOUT]

    def skip_configure(self, buildscript, last_phase):
        # don't skip this stage if we got here from one of the
        # following phases:
        if last_phase in [self.PHASE_FORCE_CHECKOUT,
                          self.PHASE_CLEAN,
                          self.PHASE_BUILD,
                          self.PHASE_INSTALL]:
            return False

        if buildscript.config.alwaysautogen:
            return False

        srcdir = self.get_srcdir(buildscript)
        builddir = self.get_builddir(buildscript)
        meson_marker_path = os.path.join(builddir, 'meson-private', 'coredata.dat')
        if not os.path.exists(meson_marker_path):
            return False

        return True

    def do_clean(self, buildscript):
        buildscript.set_action(_('Cleaning'), self)
        self.ninja(buildscript, 'clean')
    do_clean.depends = [PHASE_CONFIGURE]
    do_clean.error_phases = [PHASE_FORCE_CHECKOUT, PHASE_CONFIGURE]

    def do_build(self, buildscript):
        buildscript.set_action(_('Building'), self)
        self.ninja(buildscript)
    do_build.depends = [PHASE_CONFIGURE]
    do_build.error_phases = [PHASE_FORCE_CHECKOUT, PHASE_CONFIGURE,
            PHASE_CLEAN]

    def do_check(self, buildscript):
        buildscript.set_action(_('Checking'), self)
        self.ninja(buildscript, 'test')
    do_check.depends = [PHASE_BUILD]
    do_check.error_phases = [PHASE_FORCE_CHECKOUT, PHASE_CONFIGURE]

    def do_dist(self, buildscript):
        buildscript.set_action(_('Creating tarball for'), self)
        self.ninja(buildscript, 'dist')
    do_dist.depends = [PHASE_CONFIGURE]
    do_dist.error_phases = [PHASE_FORCE_CHECKOUT, PHASE_CONFIGURE]

    def skip_install(self, buildscript, last_phase):
        return self.config.noinstall or self.skip_install_phase

    def do_install(self, buildscript):
        buildscript.set_action(_('Installing'), self)
        destdir = self.prepare_installroot(buildscript)
        self.ninja(buildscript, 'install', env={'DESTDIR': destdir})
        self.process_install(buildscript, self.get_revision())
    do_install.depends = [PHASE_BUILD]

    def xml_tag_and_attrs(self):
        return 'meson', [('id', 'name', None),
                         ('skip-install', 'skip_install_phase', False)]


def parse_meson(node, config, uri, repositories, default_repo):
    instance = MesonModule.parse_from_xml(node, config, uri, repositories, default_repo)

    instance.dependencies += ['meson', instance.get_ninjacmd(config)]

    instance.mesonargs = collect_args(instance, node, 'mesonargs')
    instance.ninjaargs = collect_args(instance, node, 'ninjaargs')

    if node.hasAttribute('skip-install'):
        skip_install = node.getAttribute('skip-install')
        if skip_install.lower() in ('true', 'yes'):
            instance.skip_install_phase = True
        else:
            instance.skip_install_phase = False

    return instance

register_module_type('meson', parse_meson)



# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2008  Frederic Peters
#
#   build.py: custom logs pages
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

try:
    import html5lib
except ImportError:
    html5lib = None

try:
    import elementtree.ElementTree as ET
except ImportError:
    try:
        import xml.etree.ElementTree as ET
    except ImportError:
        ET = None

from buildbot.status.web.logs import LogsResource, HtmlResource, NoResource, IHTMLLog

from zope.interface import implements
from twisted.python import components


from twisted.web.resource import Resource
from twisted.web.error import NoResource

from buildbot import interfaces
from buildbot.status import builder
from buildbot.status.web.base import IHTMLLog, HtmlResource
from twisted.web.util import Redirect


class HTMLLog(HtmlResource):
    implements(IHTMLLog)

    def __init__(self, original):
        Resource.__init__(self)
        self.original = original

        parser = html5lib.HTMLParser()
        c = parser.parse(self.original.html)
        if c.childNodes[0].type == 3: # doctype
            c = c.childNodes[1]
        if c.attributes.get('xmlns'):
            del c.attributes['xmlns']
        self.x = ET.fromstring(c.toxml())

    def getTitle(self, request):
        status = self.getStatus(request)
        p = status.getProjectName()
        title = self.x.find('head/title')
        if title is not None:
            return '%s: %s' % (p, title.text)
        return p

    def body(self, request):
        return '\n'.join([ET.tostring(x) for x in self.x.find('body').getchildren()])


class JhLogsResource(LogsResource):
    def getChild(self, path, req):
        if path == '':
            return Redirect('..')
        for log in self.step_status.getLogs():
            if path == log.getName():
                if log.hasContents():
                    if html5lib and ET and hasattr(log, 'html'):
                        return HTMLLog(log)
                    return IHTMLLog(interfaces.IStatusLog(log))
                return NoResource("Empty Log '%s'" % path)
        return HtmlResource.getChild(self, path, req)




# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2001-2006  James Henstridge
# Copyright (C) 2010  Marc-Andre Lureau <marcandre.lureau@gmail.com>
#
#   fossil.py: some code to handle various Fossil operations
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

__all__ = []
__metaclass__ = type

import os
import sys
import urlparse
from subprocess import Popen, PIPE

from jhbuild.errors import FatalError, CommandError
from jhbuild.versioncontrol import Repository, Branch, register_repo_type
from jhbuild.commands.sanitycheck import inpath

class FossilRepository(Repository):
    """A class representing a Fossil repository."""

    init_xml_attrs = ['href']

    def __init__(self, config, name, href):
        Repository.__init__(self, config, name)
        # allow user to adjust location of branch.
        self.href = config.repos.get(name, href)

    branch_xml_attrs = ['module', 'checkoutdir']

    def branch(self, name, module=None, checkoutdir=None):
        if name in self.config.branches:
            module = self.config.branches[name]
            if not module:
                raise FatalError(_('branch for %(name)s has wrong override, check your %(filename)s') % \
                                   {'name'     : name,
                                    'filename' : self.config.filename})
        else:
            if module is None:
                module = name
            module = urlparse.urljoin(self.href, module)
        return FossilBranch(self, module, checkoutdir)


class FossilBranch(Branch):
    """A class representing a Fossil branch."""

    def srcdir(self):
        if self.checkoutdir:
            return os.path.join(self.checkoutroot, self.checkoutdir)
        else:
            return os.path.join(self.checkoutroot,
                                os.path.basename(self.module))

    srcdir = property(srcdir)

    def repositoryfile(self):
        return os.path.join(self.checkoutroot,
                            os.path.basename(self.checkoutdir)  + '.fossil')

    repositoryfile = property(repositoryfile)

    def branchname(self):
        return None

    branchname = property(branchname)

    def _checkout(self, buildscript):
        if self.config.sticky_date:
            raise FatalError(_('date based checkout not yet supported\n'))

        if not os.path.exists(self.repositoryfile):
            cmd = ['fossil', 'clone', self.module, self.repositoryfile]
            buildscript.execute(cmd, 'fossil', cwd=self.checkoutroot)

        os.mkdir(self.srcdir)

        cmd = ['fossil', 'open', self.repositoryfile]
        buildscript.execute(cmd, 'fossil', cwd=self.srcdir)

    def _update(self, buildscript):
        if self.config.sticky_date:
            raise FatalError(_('date based checkout not yet supported\n'))

        cmd = ['fossil', 'pull', self.module]
        buildscript.execute(cmd, 'fossil', cwd=self.srcdir)

        cmd = ['fossil', 'update']
        buildscript.execute(cmd, 'fossil', cwd=self.srcdir)

    def checkout(self, buildscript):
        if not inpath('fossil', os.environ['PATH'].split(os.pathsep)):
            raise CommandError(_('%s not found') % 'fossil')
        Branch.checkout(self, buildscript)

    def get_revision_id(self):
        import re

        try:
            infos = Popen(['fossil', 'info'], stdout=PIPE, cwd=self.srcdir)
        except OSError as e:
            raise CommandError(str(e))
        infos = infos.stdout.read().strip()
        return re.search(r"checkout: +(\w+)", infos).group(1)

    def tree_id(self):
        if not os.path.exists(self.srcdir):
            return None
        return self.get_revision_id()

register_repo_type('fossil', FossilRepository)



# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2012  Craig Keogh
#
#   systemm.py: system module support code.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from jhbuild.versioncontrol import Branch, Repository, register_repo_type
from jhbuild.utils.sxml import sxml

class SystemRepository(Repository):

    branch_xml_attrs = ['version']

    def branch(self, name, version = None):
        instance = SystemBranch(self, version)
        return instance

    def to_sxml(self):
        return [sxml.repository(type='system', name=self.name)]

class SystemBranch(Branch):

    def __init__(self, repository, version):
        Branch.__init__(self, repository, module = None, checkoutdir = None)
        self.version = version

    def branchname(self):
        return self.version
    branchname = property(branchname)

    def to_sxml(self):
        return ([sxml.branch(module=self.module,
                             repo=self.repository,
                             version=self.version)])


register_repo_type('system', SystemRepository)



# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2014 Canonical Limited
#
#  sysid.py: identify the system that we are running on
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the licence, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import sys
import subprocess
import ast

sys_id = None
sys_name = None
default_conditions = None

def read_os_release():
    global default_conditions
    global sys_name
    global sys_id

    release_file = None

    try:
        release_file = open('/etc/os-release')
    except:
        try:
            release_file = open('/usr/lib/os-release')
        except:
            return False

    fields = {}
    for line in release_file:
        line = line.strip()

        if not line or line.startswith('#'):
            continue

        if '=' not in line:
            continue

        field, _, value = line.partition('=')

        if value.startswith("'") or value.startswith('"'):
            try:
                value = ast.literal_eval(value)
            except:
                continue

        fields[field] = value

    if 'ID' not in fields or 'VERSION_ID' not in fields:
        return False

    sys_id = fields['ID'] + '-' + fields['VERSION_ID']

    if 'NAME' in fields and 'VERSION' in fields:
        sys_name = fields['NAME'] + ' ' + fields['VERSION']
    else:
        # fall back
        sys_name = fields['ID'] + ' ' + fields['VERSION_ID']

    default_conditions.add(fields['ID'])

    if 'ID_LIKE' in fields:
        default_conditions.update(fields['ID_LIKE'].split(' '))

    return True

def get_macos_info():
    global sys_name
    global sys_id

    try:
        ver = subprocess.check_output('sw_vers -productVersion')

        sys_name = 'Mac OS X ' + ver
        sys_id = 'macos-' + ver

        return True

    except:
        return False

def get_freebsd_info():
    global sys_name
    global sys_id

    try:
        ver = subprocess.check_output('freebsd-version').strip()

        sys_name = 'FreeBSD ' + ver
        return True
    except:
        pass

    try:
        ver = subprocess.check_output(['uname', '-r']).strip()

        sys_name = 'FreeBSD ' + ver
        return True
    except:
        return False

def ensure_loaded():
    global default_conditions
    global sys_name
    global sys_id

    if sys_id is not None:
        return

    # the default conditions set.  We determine which set to used based on
    # the first item in the list which is a prefix of 'sys.platform', which
    # is a name like 'linux2', 'darwin', 'freebsd10', etc.
    #
    # if we watch to match (eg 'freebsd10' more closely than other versions
    # of 'freebsd') then we just need to make sure the more-specific one
    # comes first in the list
    conditions_sets = [
            ('linux', ['linux', 'wayland', 'udev', 'udev-hwdb', 'evdev', 'x11',
             'systemd', 'gnu-elf']),
            ('freebsd', ['freebsd', 'wayland', 'udev', 'x11', 'bsd', 'gnu-elf',
             'gmake']),
            ('darwin', ['darwin', 'macos', 'quartz']),

            # this must be left here so that at least one will be found
            ('', ['x11'])
        ]

    for prefix, flags in conditions_sets:
        if sys.platform.startswith(prefix):
            default_conditions = set(flags)
            break

    # our first choice is to use os-release info
    if read_os_release():
        return

    # but failing that, fall back to using sys.platform
    sys_id = sys.platform

    if sys_id.startswith('linux'):
        sys_name = "Unknown Linux Distribution (no 'os-release' file)"

    elif sys_id.startswith('freebsd'):
        if not get_freebsd_info():
            sys_name = 'FreeBSD (%s)' % (sys_id)

    elif sys_id.startswith('macos'):
        if not get_macos_info():
            sys_name = 'Mac OS X (unknown version)'

    else:
        sys_id = sys.platform
        sys_name = sys.platform

def get_id():
    ensure_loaded()

    return sys_id

def get_pretty_name():
    ensure_loaded()

    return sys_name

def get_default_conditions():
    ensure_loaded()

    return default_conditions
